# -*- coding: utf-8 -*-


import pytest

from aerospike_helpers.expressions import (
    Eq,
    IntAnd,
    IntArithmeticRightShift,
    IntBin,
    IntCount,
    IntLeftScan,
    IntLeftShift,
    IntNot,
    IntOr,
    IntRightScan,
    IntRightShift,
    IntXOr,
)
from aerospike import exception as e
from .test_base_class import TestBaseClass


class TestExpressionsBitOps(TestBaseClass):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        if self.server_version < [5, 6]:
            pytest.mark.xfail(reason="Servers older than 5.6 do not support bitwise operator expressions.")
            pytest.xfail()

        self.test_ns = "test"
        self.test_set = "demo"

        self.key = ("test", "demo", 5)
        self.rec = {
            "1bin": 0xFFFF,  # (255 * 2), 65535, FFFF, 1111111111111111
            "0bin": 0x0000,  # (0 * 2),   0,     0,    0000000000000000
            "10bin": 0xAAAA,  # (170 * 2), 43690, AAAA, 1010101010101010
        }
        self.as_connection.put(self.key, self.rec)

        def teardown():
            as_connection.remove(self.key)

        request.addfinalizer(teardown)

    def verify_expression(self, expr, expected):
        _, _, res = self.as_connection.get(self.key, policy={"expressions": expr})
        assert res == self.rec

    def verify_expression_neg(self, expr, expected):
        with pytest.raises(expected):
            _, _, res = self.as_connection.get(self.key, policy={"expressions": expr})

    @pytest.mark.parametrize(
        "bin, val, check",
        [
            (IntBin("10bin"), [0xFF00], 0xAA00),
            (IntBin("1bin"), [0x0000], 0x0000),
            (IntBin("0bin"), [0xFF00], 0x0000),
        ],
    )
    def test_int_and_pos(self, bin, val, check):
        """
        Test IntAnd expression with correct parameters.
        """
        expr = Eq(IntAnd(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [(IntBin("10bin"), [0xFF00], 0xAAF0, e.FilteredOut), (IntBin("1bin"), ["bad_arg"], 0x0000, e.InvalidRequest)],
    )
    def test_int_and_neg(self, bin, val, expected, check):
        """
        Test IntAnd expression with incorrect parameters.
        """
        expr = Eq(IntAnd(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "bin, val, check",
        [
            (IntBin("10bin"), [0xFF00], 0xFFAA),
            (IntBin("1bin"), [0x0000], 0xFFFF),
            (IntBin("0bin"), [0xFF00], 0xFF00),
        ],
    )
    def test_int_or_pos(self, bin, val, check):
        """
        Test IntOr expression with correct parameters.
        """
        expr = Eq(IntOr(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [(IntBin("10bin"), [0xFF00], 0xAAF0, e.FilteredOut), (IntBin("1bin"), ["bad_arg"], 0x0000, e.InvalidRequest)],
    )
    def test_int_or_neg(self, bin, val, expected, check):
        """
        Test IntOr expression with incorrect parameters.
        """
        expr = Eq(IntOr(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "bin, val, check",
        [
            (IntBin("10bin"), [0xFF00], 0x55AA),
            (IntBin("1bin"), [0x0000], 0xFFFF),
            (IntBin("0bin"), [0xFF00], 0xFF00),
        ],
    )
    def test_int_xor_pos(self, bin, val, check):
        """
        Test IntXOr expression with correct parameters.
        """
        expr = Eq(IntXOr(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [(IntBin("10bin"), [0xFF00], 0xAAF0, e.FilteredOut), (IntBin("1bin"), ["bad_arg"], 0x0000, e.InvalidRequest)],
    )
    def test_int_xor_neg(self, bin, val, expected, check):
        """
        Test IntXOr expression with incorrect parameters.
        """
        expr = Eq(IntXOr(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "bin, check",
        [
            (IntBin("10bin"), ~0xAAAA),
            (IntBin("1bin"), ~0xFFFF),
            (IntBin("0bin"), ~0x0000),
        ],
    )
    def test_int_not_pos(self, bin, check):
        """
        Test IntNot expression with correct parameters.
        """
        expr = Eq(IntNot(bin), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, check, expected", [(IntBin("10bin"), 0xAAF0, e.FilteredOut), ("bad_arg", 0x0000, e.InvalidRequest)]
    )
    def test_int_not_neg(self, bin, expected, check):
        """
        Test IntNot expression with incorrect parameters.
        """
        expr = Eq(IntNot(bin), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "bin, val, check",
        [
            (IntBin("10bin"), [2], 0xAAAA << 2),
            (IntBin("1bin"), [4], 0xFFFF << 4),
            (IntBin("0bin"), [1], 0x0000 << 1),
        ],
    )
    def test_int_left_shift_pos(self, bin, val, check):
        """
        Test IntLeftShift expression with correct parameters.
        """
        expr = Eq(IntLeftShift(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [
            (IntBin("10bin"), [2], 0xAAF0, e.FilteredOut),
            # InvalidRequest because shift is not type checked (expression).
            (IntBin("1bin"), ["bad_arg"], 0x0000, e.InvalidRequest),
        ],
    )
    def test_int_left_shift_neg(self, bin, val, expected, check):
        """
        Test IntLeftShift expression with incorrect parameters.
        """
        expr = Eq(IntLeftShift(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "bin, val, check",
        [
            (IntBin("10bin"), [2], 0xAAAA >> 2),
            (IntBin("1bin"), [4], 0xFFFF >> 4),
            (IntBin("0bin"), [1], 0x0000 >> 1),
            # (-255, [1], -255 >> 1), TODO add case distinct from arithmetic shift
        ],
    )
    def test_int_right_shift_pos(self, bin, val, check):
        """
        Test IntRightShift expression with correct parameters.
        """
        expr = Eq(IntRightShift(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [
            (IntBin("10bin"), [2], 0xAAF0, e.FilteredOut),
            # InvalidRequest because shift is not type checked (expression).
            (IntBin("1bin"), ["bad_arg"], 0xFFFF, e.InvalidRequest),
        ],
    )
    def test_int_right_shift_neg(self, bin, val, expected, check):
        """
        Test IntRightShift expression with incorrect parameters.
        """
        expr = Eq(IntRightShift(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "bin, val, check",
        [
            (IntBin("10bin"), [2], 0xAAAA >> 2),
            (IntBin("1bin"), [4], 0xFFFF >> 4),
            (IntBin("0bin"), [1], 0x0000 >> 1),
            # TODO add case distinct from logical shift
        ],
    )
    def test_int_right_arithmetic_shift_pos(self, bin, val, check):
        """
        Test IntArithmeticRightShift expression with correct parameters.
        """
        expr = Eq(IntArithmeticRightShift(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [
            (IntBin("10bin"), [2], 0xAAF0, e.FilteredOut),
            # InvalidRequest because shift is not type checked (expression).
            (IntBin("1bin"), ["bad_arg"], 0xFFFF, e.InvalidRequest),
        ],
    )
    def test_int_right_arithmetic_shift_neg(self, bin, val, expected, check):
        """
        Test IntArithmeticRightShift expression with incorrect parameters.
        """
        expr = Eq(IntArithmeticRightShift(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "bin, check",
        [
            (IntBin("10bin"), 8),
            (IntBin("1bin"), 16),
            (IntBin("0bin"), 0),
        ],
    )
    def test_int_count_pos(self, bin, check):
        """
        Test IntCount expression with correct parameters.
        """
        expr = Eq(IntCount(bin), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, check, expected", [(IntBin("10bin"), 10, e.FilteredOut), ("bad_arg", 0, e.InvalidRequest)]
    )
    def test_int_count_neg(self, bin, expected, check):
        """
        Test IntCount expression with incorrect parameters.
        """
        expr = Eq(IntCount(bin), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "bin, val, check",
        [  # NOTE: Aerospike ints are 64bit.
            (1, [True], 63),
            (IntBin("1bin"), [True], 48),
            (IntBin("0bin"), [False], 0),
        ],
    )
    def test_int_left_scan_pos(self, bin, val, check):
        """
        Test IntLeftScan expression with correct parameters.
        """
        expr = Eq(IntLeftScan(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [(IntBin("0bin"), [True], 0, e.FilteredOut), (IntBin("1bin"), ["bad_arg"], 0x0000, e.InvalidRequest)],
    )
    def test_int_left_scan_neg(self, bin, val, expected, check):
        """
        Test IntLeftScan expression with incorrect parameters.
        """
        expr = Eq(IntLeftScan(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)

    @pytest.mark.parametrize(
        "bin, val, check",
        [  # NOTE: Aerospike ints are 64bit.
            (1, [True], 63),
            (IntBin("1bin"), [True], 63),
            (IntBin("0bin"), [False], 63),
        ],
    )
    def test_int_right_scan_pos(self, bin, val, check):
        """
        Test IntRightScan expression with correct parameters.
        """
        expr = Eq(IntRightScan(bin, *val), check).compile()

        self.verify_expression(expr, self.rec)

    @pytest.mark.parametrize(
        "bin, val, check, expected",
        [(IntBin("0bin"), [True], 0, e.FilteredOut), (IntBin("1bin"), ["bad_arg"], 0x0000, e.InvalidRequest)],
    )
    def test_int_right_scan_neg(self, bin, val, expected, check):
        """
        Test IntRightScan expression with incorrect parameters.
        """
        expr = Eq(IntRightScan(bin, *val), check).compile()

        self.verify_expression_neg(expr, expected)



#
# Copyright 2008-2021 Aerospike, Inc.
#
# Portions may be licensed to Aerospike, Inc. under one or more contributor
# license agreements.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

#
#    Client Errors
#
#
#    Synchronous connection error.
#
AEROSPIKE_ERR_CONNECTION = -10

#
#    Node invalid or could not be found.
#
AEROSPIKE_ERR_TLS_ERROR = -9

#
#    Node invalid or could not be found.
#
AEROSPIKE_ERR_INVALID_NODE = -8

#
#    Asynchronous connection error.
#
AEROSPIKE_ERR_NO_MORE_CONNECTIONS = -7

#
#    Asynchronous connection error.
#
AEROSPIKE_ERR_ASYNC_CONNECTION = -6

#
#    Query or scan was aborted in user's callback.
#
AEROSPIKE_ERR_CLIENT_ABORT = -5

#
#    Host name could not be found in DNS lookup.
#
AEROSPIKE_ERR_INVALID_HOST = -4

#
#    No more records available when parsing batch, scan or query records.
#
AEROSPIKE_NO_MORE_RECORDS = -3

#
#    Invalid client API parameter.
#
AEROSPIKE_ERR_PARAM = -2

#
#    Generic client API usage error.
#
AEROSPIKE_ERR_CLIENT = -1

#
#    Deprecated.  Generic client error.  Keep for legacy reasons.
#
AEROSPIKE_ERR = -1

#
#    Success
#

#
#    Generic success.
#
AEROSPIKE_OK = 0

#
#    Server Errors
#

#
#    Generic error returned by server.
#
AEROSPIKE_ERR_SERVER = 1

#
#    Record does not exist in database. May be returned by read, or write
#    with policy AS_POLICY_EXISTS_UPDATE.
#
AEROSPIKE_ERR_RECORD_NOT_FOUND = 2

#
#    Generation of record in database does not satisfy write policy.
#
AEROSPIKE_ERR_RECORD_GENERATION = 3

#
#    Request protocol invalid, or invalid protocol field.
#
AEROSPIKE_ERR_REQUEST_INVALID = 4

#
#    Record already exists. May be returned by write with policy
#    AS_POLICY_EXISTS_CREATE.
#
AEROSPIKE_ERR_RECORD_EXISTS = 5

#
#    Bin already exists.
#
AEROSPIKE_ERR_BIN_EXISTS = 6

#
#    A cluster state change occurred during the request. This may also be
#    returned by scan operations with the fail_on_cluster_change flag set.
#
AEROSPIKE_ERR_CLUSTER_CHANGE = 7

#
# or storage device space
#    reserved for the specified namespace.
#
AEROSPIKE_ERR_SERVER_FULL = 8

#
#    Request timed out.  Can be triggered by client or server.
#
AEROSPIKE_ERR_TIMEOUT = 9

#
#    XDR is not available for the cluster.
#
AEROSPIKE_ERR_NO_XDR = 10

#
#    Generic cluster discovery & connection error.
#
AEROSPIKE_ERR_CLUSTER = 11

#
#    Bin modification operation can't be done on an existing bin due to its
#    value type.
#
AEROSPIKE_ERR_BIN_INCOMPATIBLE_TYPE = 12

#
#    Record being (re-)written can't fit in a storage write block.
#
AEROSPIKE_ERR_RECORD_TOO_BIG = 13

#
#    Too may concurrent requests for one record - a "hot-key" situation.
#
AEROSPIKE_ERR_RECORD_BUSY = 14

#
#    Scan aborted by user.
#
AEROSPIKE_ERR_SCAN_ABORTED = 15

#
#    Sometimes our doc, or our customers wishes, get ahead of us.  We may have
#    processed something that the server is not ready for (unsupported feature).
#
AEROSPIKE_ERR_UNSUPPORTED_FEATURE = 16

#
#    Bin-level replace-only supported on server but not on client.
#
AEROSPIKE_ERR_BIN_NOT_FOUND = 17

#
#    The server node's storage device(s) can't keep up with the write load.
#
AEROSPIKE_ERR_DEVICE_OVERLOAD = 18

#
#    Record key sent with transaction did not match key stored on server.
#
AEROSPIKE_ERR_RECORD_KEY_MISMATCH = 19

#
#    Namespace in request not found on server.
#
AEROSPIKE_ERR_NAMESPACE_NOT_FOUND = 20

#
#    Sent too-long bin name (should be impossible in this client) or exceeded
#    namespace's bin name quota.
#
AEROSPIKE_ERR_BIN_NAME = 21

#
#    Operation not allowed at this time.
#
AEROSPIKE_ERR_FAIL_FORBIDDEN = 22

AEROSPIKE_ERR_FAIL_ELEMENT_NOT_FOUND = 23

AEROSPIKE_ERR_FAIL_ELEMENT_EXISTS = 24

#
#    There are no more records left for query.
#
AEROSPIKE_QUERY_END = 50

#
#    Security functionality not supported by connected server.
#
AEROSPIKE_SECURITY_NOT_SUPPORTED = 51

#
#    Security functionality not enabled by connected server.
#
AEROSPIKE_SECURITY_NOT_ENABLED = 52

#
#    Security type not supported by connected server.
#
AEROSPIKE_SECURITY_SCHEME_NOT_SUPPORTED = 53

#
#    Administration command is invalid.
#
AEROSPIKE_INVALID_COMMAND = 54

#
#    Administration field is invalid.
#
AEROSPIKE_INVALID_FIELD = 55

#
#    Security protocol not followed.
#
AEROSPIKE_ILLEGAL_STATE = 56

#
#    User name is invalid.
#
AEROSPIKE_INVALID_USER = 60

#
#    User was previously created.
#
AEROSPIKE_USER_ALREADY_EXISTS = 61

#
#    Password is invalid.
#
AEROSPIKE_INVALID_PASSWORD = 62

#
#    Password has expired.
#
AEROSPIKE_EXPIRED_PASSWORD = 63

#
#    Forbidden password (e.g. recently used)
#
AEROSPIKE_FORBIDDEN_PASSWORD = 64

#
#    Security credential is invalid.
#
AEROSPIKE_INVALID_CREDENTIAL = 65

#
#    Role name is invalid.
#
AEROSPIKE_INVALID_ROLE = 70

#
#    Role already exists.
#
AEROSPIKE_ROLE_ALREADY_EXISTS = 71

#
#    Privilege is invalid.
#
AEROSPIKE_INVALID_PRIVILEGE = 72

#
#    User must be authentication before performing database operations.
#
AEROSPIKE_NOT_AUTHENTICATED = 80

#
#    User does not possess the required role to perform the database operation.
#
AEROSPIKE_ROLE_VIOLATION = 81

#
#    Generic UDF error.
#
AEROSPIKE_ERR_UDF = 100

#
#    The requested item in a large collection was not found.
#
AEROSPIKE_ERR_LARGE_ITEM_NOT_FOUND = 125

#
# Batch functionality has been disabled.
#
AEROSPIKE_ERR_BATCH_DISABLED = 150

#
# Batch max requests have been exceeded.
#
AEROSPIKE_ERR_BATCH_MAX_REQUESTS_EXCEEDED = 151

#
# All batch queues are full.
#
AEROSPIKE_ERR_BATCH_QUEUES_FULL = 152

#
# Unsupported GeoJSON
#
AEROSPIKE_ERR_GEO_INVALID_GEOJSON = 160

#
#    Index found.
#
AEROSPIKE_ERR_INDEX_FOUND = 200

#
#    Index not found
#
AEROSPIKE_ERR_INDEX_NOT_FOUND = 201

#
#    Index is out of memory
#
AEROSPIKE_ERR_INDEX_OOM = 202

#
#    Unable to read the index.
#
AEROSPIKE_ERR_INDEX_NOT_READABLE = 203

#
#    Generic secondary index error.
#
AEROSPIKE_ERR_INDEX = 204

#
#    Index name is too long.
#
AEROSPIKE_ERR_INDEX_NAME_MAXLEN = 205

#
#    System already has maximum allowed indices.
#
AEROSPIKE_ERR_INDEX_MAXCOUNT = 206

#
#    Query was aborted.
#
AEROSPIKE_ERR_QUERY_ABORTED = 210

#
#    Query processing queue is full.
#
AEROSPIKE_ERR_QUERY_QUEUE_FULL = 211

#
#    Secondary index query timed out on server.
#
AEROSPIKE_ERR_QUERY_TIMEOUT = 212

#
#    Generic query error.
#
AEROSPIKE_ERR_QUERY = 213

#
#    UDF OPERATIONS
#

#
#    UDF does not exist.
#
AEROSPIKE_ERR_UDF_NOT_FOUND = 1301
#
#    LUA file does not exist.
#
AEROSPIKE_ERR_LUA_FILE_NOT_FOUND = 1302

#
#    Large Data Type (LDT) OPERATIONS
#

#
AEROSPIKE_ERR_LDT_INTERNAL = 1400

#
AEROSPIKE_ERR_LDT_NOT_FOUND = 1401

#
AEROSPIKE_ERR_LDT_UNIQUE_KEY = 1402

#
AEROSPIKE_ERR_LDT_INSERT = 1403

#
AEROSPIKE_ERR_LDT_SEARCH = 1404

#
AEROSPIKE_ERR_LDT_DELETE = 1405


#
AEROSPIKE_ERR_LDT_INPUT_PARM = 1409

# -------------------------------------------------

#
AEROSPIKE_ERR_LDT_TYPE_MISMATCH = 1410

#
AEROSPIKE_ERR_LDT_NULL_BIN_NAME = 1411

#
AEROSPIKE_ERR_LDT_BIN_NAME_NOT_STRING = 1412

#
AEROSPIKE_ERR_LDT_BIN_NAME_TOO_LONG = 1413

#
AEROSPIKE_ERR_LDT_TOO_MANY_OPEN_SUBRECS = 1414

#
AEROSPIKE_ERR_LDT_TOP_REC_NOT_FOUND = 1415

#
AEROSPIKE_ERR_LDT_SUB_REC_NOT_FOUND = 1416

#
AEROSPIKE_ERR_LDT_BIN_DOES_NOT_EXIST = 1417

#
AEROSPIKE_ERR_LDT_BIN_ALREADY_EXISTS = 1418

#
AEROSPIKE_ERR_LDT_BIN_DAMAGED = 1419

# -------------------------------------------------

#
AEROSPIKE_ERR_LDT_SUBREC_POOL_DAMAGED = 1420

#
AEROSPIKE_ERR_LDT_SUBREC_DAMAGED = 1421

#
AEROSPIKE_ERR_LDT_SUBREC_OPEN = 1422

#
AEROSPIKE_ERR_LDT_SUBREC_UPDATE = 1423

#
AEROSPIKE_ERR_LDT_SUBREC_CREATE = 1424

#
AEROSPIKE_ERR_LDT_SUBREC_DELETE = 1425

#
AEROSPIKE_ERR_LDT_SUBREC_CLOSE = 1426

#
AEROSPIKE_ERR_LDT_TOPREC_UPDATE = 1427

#
AEROSPIKE_ERR_LDT_TOPREC_CREATE = 1428

# -------------------------------------------------

#
AEROSPIKE_ERR_LDT_FILTER_FUNCTION_BAD = 1430

#
AEROSPIKE_ERR_LDT_FILTER_FUNCTION_NOT_FOUND = 1431

#
AEROSPIKE_ERR_LDT_KEY_FUNCTION_BAD = 1432

#
AEROSPIKE_ERR_LDT_KEY_FUNCTION_NOT_FOUND = 1433

#
AEROSPIKE_ERR_LDT_TRANS_FUNCTION_BAD = 1434

#
AEROSPIKE_ERR_LDT_TRANS_FUNCTION_NOT_FOUND = 1435

#
AEROSPIKE_ERR_LDT_UNTRANS_FUNCTION_BAD = 1436

#
AEROSPIKE_ERR_LDT_UNTRANS_FUNCTION_NOT_FOUND = 1437

#
AEROSPIKE_ERR_LDT_USER_MODULE_BAD = 1438

#
AEROSPIKE_ERR_LDT_USER_MODULE_NOT_FOUND = 1439



# -*- coding: utf-8 -*-
import pytest
import time

from aerospike_helpers import expressions as exp
from aerospike import exception as e

import aerospike


def seconds_to_nanos(num):
    """
    converter seconds to nanoseconds
    """
    return int(num) * (10**9)


# GeoConstants
geo_object1 = aerospike.GeoJSON({"type": "AeroCircle", "coordinates": [[-122.0, 37.5], 1000]})

geo_object2 = aerospike.GeoJSON({"type": "AeroCircle", "coordinates": [[-132.0, 37.5], 1000]})

geo_point1 = aerospike.GeoJSON({"coordinates": [-122.0, 37.5], "type": "Point"})

geo_point2 = aerospike.GeoJSON({"coordinates": [-132.0, 37.5], "type": "Point"})


def assert_each_record(records, check_func, *args):
    for record in records:
        assert check_func(record, *args)


def assert_each_record_bins(records, check_func):
    for _, _, bins in records:
        assert check_func(bins)


@pytest.fixture(scope="class")
def clean_test_demo_namespace(as_connection):
    names = ["Alice", "Bob", "John", "Jane"]
    for i in range(100):
        key = "test", "demo", i
        record = {
            "name": names[i % 4],
            "positive_i": i,
            "i_mod_5": i % 5,
            "i_mod_10": i % 10,
            # This is a list: [i, i + 1, i +2 , i+3, i+4, i+5]
            "plus_five_l": list(range(i, i + 6)),
        }
        as_connection.put(key, record)

        #  For list tests
        flist1 = ["Alice", "Bob", "John"]
        flist2 = ["Alice", "Bob", "Jane"]
        flist3 = ["John", "Jane"]
        key1 = "test", "demo2", "f1"
        key2 = "test", "demo2", "f2"
        key3 = "test", "demo2", "f3"

        as_connection.put(key1, {"slist": flist1})
        as_connection.put(key2, {"slist": flist2})
        as_connection.put(key3, {"slist": flist3})

        #  For map tests:
        flist1 = ["Alice", "Bob", "John"]
        flist2 = ["Alice", "Bob", "Jane"]
        flist3 = ["John", "Jane"]

        # dictionary: {'Alice': 5, 'Bob': '3'...}
        map1 = dict([(name, len(name)) for name in flist1])
        map2 = dict([(name, len(name)) for name in flist2])
        map3 = dict([(name, len(name)) for name in flist3])

        key1 = "test", "demo3", "f1"
        key2 = "test", "demo3", "f2"
        key3 = "test", "demo3", "f3"

        as_connection.put(key1, {"map": map1})
        as_connection.put(key2, {"map": map2})
        as_connection.put(key3, {"map": map3})

        geok1 = "test", "geo", 1
        geok2 = "test", "geo", 2

        georec1 = {"id": 1, "point": geo_point1, "region": geo_object1, "geolist": [geo_point1]}

        georec2 = {"id": 2, "point": geo_point2, "region": geo_object2, "geolist": [geo_point2]}

        as_connection.put(geok1, georec1)
        as_connection.put(geok2, georec2)

    yield

    as_connection.truncate("test", None, 0)


@pytest.mark.usefixtures("clean_test_demo_namespace")
class TestQueryExpressions(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        self.del_keys = []
        self.query = as_connection.query("test", "demo")

    def test_integer_equals(self):
        expr = exp.Eq(exp.IntBin("positive_i"), 5)
        results = self.query.results(policy={"expressions": expr.compile()})
        assert len(results) == 1
        assert_each_record_bins(results, lambda b: b["positive_i"] == 5)

    def test_integer_greater(self):
        expr = exp.GT(exp.IntBin("positive_i"), 49)
        results = self.query.results(policy={"expressions": expr.compile()})
        assert len(results) == 50
        assert_each_record_bins(results, lambda b: b["positive_i"] > 49)

    def test_integer_greatereq(self):
        expr = exp.GE(exp.IntBin("positive_i"), 49)
        results = self.query.results(policy={"expressions": expr.compile()})
        assert len(results) == 51
        assert_each_record_bins(results, lambda b: b["positive_i"] >= 49)

    def test_integer_less(self):
        expr = exp.LT(exp.IntBin("positive_i"), 10)
        results = self.query.results(policy={"expressions": expr.compile()})
        assert len(results) == 10
        assert_each_record_bins(results, lambda b: b["positive_i"] < 10)

    def test_integer_lesseq(self):
        expr = exp.LE(exp.IntBin("positive_i"), 10)
        results = self.query.results(policy={"expressions": expr.compile()})
        assert len(results) == 11
        assert_each_record_bins(results, lambda b: b["positive_i"] <= 10)

    # TBD What calculation is the server doing here, I don't get it right now.

    def test_digest_modulo(self):
        # Count of digests whose last byte is < 128
        less_than_128 = 0
        expected_ids = set([])
        for i in range(100):
            key = "test", "demo", i
            if aerospike.calc_digest(*key)[-1] < 128:
                expected_ids.add(i)
                less_than_128 += 1

        expr = exp.LT(exp.DigestMod(256), 128)

        results = self.query.results(policy={"expressions": expr.compile()})
        assert len(results) == less_than_128

    def test_string_equal(self):
        expr = exp.Eq(exp.StrBin("name"), "Alice")
        results = self.query.results(policy={"expressions": expr.compile()})
        assert len(results) == 25
        assert_each_record_bins(results, lambda b: b["name"] == "Alice")

    def test_string_unequal(self):
        expr = exp.NE(exp.StrBin("name"), "Alice")
        results = self.query.results(policy={"expressions": expr.compile()})
        assert len(results) == 75
        assert_each_record_bins(results, lambda b: b["name"] != "Alice")

    # TODO GEO BIN WITHIN CONTAINS

    def test_geo_within(self):
        expr = exp.CmpGeo(exp.GeoBin("point"), geo_object1)

        query = self.as_connection.query("test", "geo")
        results = query.results(policy={"expressions": expr.compile()})
        assert len(results) == 1
        assert results[0][2]["id"] == 1

    def test_geo_contains(self):
        expr = exp.CmpGeo(geo_point2, exp.GeoBin("region"))

        query = self.as_connection.query("test", "geo")
        results = query.results(policy={"expressions": expr.compile()})
        assert len(results) == 1
        assert results[0][2]["id"] == 2

    def test_geo_contains2(self):
        expr = exp.CmpGeo(exp.GeoBin("point"), exp.GeoBin("region"))

        query = self.as_connection.query("test", "geo")
        results = query.results(policy={"expressions": expr.compile()})
        assert len(results) == 2

    def test_not(self):
        expr = exp.Not(exp.Eq(exp.IntBin("positive_i"), 5))
        results = self.query.results(policy={"expressions": expr.compile()})

        assert len(results) == 99
        assert_each_record_bins(results, lambda b: b["positive_i"] != 5)

    def test_or(self):
        expr = exp.Or(exp.Eq(exp.IntBin("positive_i"), 5), exp.Eq(exp.IntBin("positive_i"), 10))
        results = self.query.results(policy={"expressions": expr.compile()})
        assert len(results) == 2
        assert_each_record_bins(results, lambda b: b["positive_i"] in (5, 10))

    def test_and(self):
        expr = exp.And(exp.GT(exp.IntBin("positive_i"), 10), exp.LT(exp.IntBin("positive_i"), 20))
        results = self.query.results(policy={"expressions": expr.compile()})
        assert len(results) == 9
        assert_each_record_bins(results, lambda b: b["positive_i"] > 10 and b["positive_i"] < 20)

    def test_string_regex(self):
        expr = exp.CmpRegex(aerospike.REGEX_ICASE, ".*O.*", exp.StrBin("name"))
        results = self.query.results(policy={"expressions": expr.compile()})
        assert len(results) == 50
        assert_each_record_bins(results, lambda b: b["name"] in ("Bob", "John"))

    # List Tests
    def test_list_or_int(self):
        expr = exp.GT(
            exp.ListGetByRank(None, aerospike.LIST_RETURN_VALUE, exp.ResultType.INTEGER, -1, "plus_five_l"), 10
        )
        results = self.query.results(policy={"expressions": expr.compile()})
        assert len(results) == 94  # This isn't true for 0,1,2,3,4,5 so 100 - 6
        assert_each_record_bins(results, lambda b: b["plus_five_l"][-1] > 10)

    def test_list_and_int(self):
        expr = exp.GT(
            exp.ListGetByRank(None, aerospike.LIST_RETURN_VALUE, exp.ResultType.INTEGER, 0, "plus_five_l"), 10
        )
        results = self.query.results(policy={"expressions": expr.compile()})
        assert len(results) == 89  # This isn't true for the first 11
        assert_each_record_bins(results, lambda b: b["plus_five_l"][-1] > 10)

    def test_list_or_str(self):
        expr = exp.GE(exp.ListGetByValue(None, aerospike.LIST_RETURN_COUNT, "Bob", "slist"), 1)
        query = self.as_connection.query("test", "demo2")
        results = query.results(policy={"expressions": expr.compile()})
        assert len(results) == 2
        assert_each_record_bins(results, lambda b: any([name == "Bob" for name in b["slist"]]))

    def test_list_and_str(self):
        expr = exp.Eq(exp.ListGetByValue(None, aerospike.LIST_RETURN_COUNT, "Bob", "slist"), 0)
        # Only one friend list without Bob
        query = self.as_connection.query("test", "demo2")
        results = query.results(policy={"expressions": expr.compile()})
        assert len(results) == 1
        assert_each_record_bins(results, lambda b: all([name != "Bob" for name in b["slist"]]))

    # Mapkey Tests
    def test_mapkey_iterate_or(self):
        expr = exp.GE(exp.MapGetByKey(None, aerospike.LIST_RETURN_COUNT, exp.ResultType.INTEGER, "Bob", "map"), 1)
        query = self.as_connection.query("test", "demo3")
        results = query.results(policy={"expressions": expr.compile()})
        assert len(results) == 2
        assert_each_record_bins(results, lambda b: any([key == "Bob" for key in b["map"]]))

    def test_mapkey_iterate_and(self):
        expr = exp.Eq(exp.MapGetByKey(None, aerospike.LIST_RETURN_COUNT, exp.ResultType.INTEGER, "Bob", "map"), 0)
        query = self.as_connection.query("test", "demo3")
        results = query.results(policy={"expressions": expr.compile()})
        assert len(results) == 1
        assert_each_record_bins(results, lambda b: all([key != "Bob" for key in b["map"]]))

    # MapValueTest
    def test_mapvalue_iterate_or(self):
        expr = exp.GE(exp.MapGetByValue(None, aerospike.LIST_RETURN_COUNT, 3, "map"), 1)
        query = self.as_connection.query("test", "demo3")
        results = query.results(policy={"expressions": expr.compile()})
        assert len(results) == 2
        assert_each_record_bins(results, lambda b: any([b["map"][key] == 3 for key in b["map"]]))

    def test_mapvalue_iterate_and(self):
        expr = exp.Eq(exp.MapGetByValue(None, aerospike.LIST_RETURN_COUNT, 3, "map"), 0)
        query = self.as_connection.query("test", "demo3")
        results = query.results(policy={"expressions": expr.compile()})
        assert len(results) == 1
        assert_each_record_bins(results, lambda b: all([b["map"][key] != 3 for key in b["map"]]))

    @pytest.mark.xfail(reason="This only works when not running data in memory")  # TODO test this on device config
    def test_rec_device_size(self):
        long_str_len = 65 * 1024
        long_str = long_str_len * "a"  # A 65K string

        # Store 5 records with a string of 65K
        for i in range(5):
            key = "test", "dev_size", i
            self.as_connection.put(key, {"string": long_str})

        # Store 3 records with a size much less than 65K
        for i in range(5, 8):
            key = "test", "dev_size", i
            self.as_connection.put(key, {"string": "short"})

        query = self.as_connection.query("test", "dev_size")
        expr = exp.GT(exp.DeviceSize(), 64 * 1024)
        results = query.results(policy={"expressions": expr.compile()})
        assert len(results) == 5
        assert_each_record_bins(
            results, lambda b: len(b["string"]) == long_str_len
        )  # This is faster than doing a string compare

    def test_rec_last_update(self):
        """
        This could fail due to clock skew
        """
        for i in range(7):
            key = "test", "lut", i
            self.as_connection.put(key, {"time": "earlier"})

        cutoff_nanos = seconds_to_nanos(int(time.time() + 2))

        time.sleep(5)  # Make sure that we wait long enough

        # Store 5 records after the cutoff
        for i in range(7, 12):
            key = "test", "lut", i
            self.as_connection.put(key, {"time": "later"})

        query = self.as_connection.query("test", "lut")

        expr = exp.LT(exp.LastUpdateTime(), cutoff_nanos)
        results = query.results(policy={"expressions": expr.compile()})
        assert len(results) == 7
        assert_each_record_bins(results, lambda b: b["time"] == "earlier")

    def test_rec_void_time(self):
        """
        This could fail due to clock skew
        """
        for i in range(7):
            key = "test", "ttl", i
            self.as_connection.put(key, {"time": "earlier"}, meta={"ttl": 100})

        # 150 second range for record TTLs should be enough, we are storing with
        # Current time + 100s and current time +5000s, so only one of the group should be found
        void_time_range_start = seconds_to_nanos(int(time.time() + 50))
        void_time_range_end = seconds_to_nanos(int(time.time() + 150))

        # Store 5 records after the cutoff
        for i in range(7, 12):
            key = "test", "ttl", i
            self.as_connection.put(key, {"time": "later"}, meta={"ttl": 1000})

        query = self.as_connection.query("test", "ttl")

        expr = exp.And(exp.GT(exp.VoidTime(), void_time_range_start), exp.LT(exp.VoidTime(), void_time_range_end))
        results = query.results(policy={"expressions": expr.compile()})
        assert len(results) == 7
        assert_each_record_bins(results, lambda b: b["time"] == "earlier")

    def test_with_eq_expr_too_big(self):
        expr = exp.Eq(1, 1 << 64)  # This needs to be over 2 ^ 63 - 1
        query = self.as_connection.query("test", "or")
        with pytest.raises(e.ParamError):
            query.results(policy={"expressions": expr.compile()})

    def test_with_invalid_predicate(self):
        """
        This passes something which isn't a predicate
        """
        expr = exp.Or(exp.Eq(1, 1), "bad_expression")
        with pytest.raises(e.InvalidRequest):
            self.query.results(policy={"expressions": expr.compile()})

    @pytest.mark.parametrize(
        "func",
        [
            exp.IntBin,
            exp.StrBin,
            exp.GeoBin,
            exp.MapBin,
            exp.ListBin,
        ],
    )
    def test_with_wrong_predicate_argument_type_expecting_str(self, func):
        """
        These functions all expect an integer argument, call with a string
        """
        expr = func(5)
        with pytest.raises(e.ParamError):
            self.query.results(policy={"expressions": expr.compile()})

    def test_with_invalid_predicate_tuple(self):
        """
        This passes something which isn't a predicate
        """
        expr = (1234, "not real")
        with pytest.raises(e.ParamError):
            self.query.results(policy={"expressions": expr})

    def test_with_empty_exp_list(self):
        """
        Pass an empty list of predicates
        """
        expr = []
        with pytest.raises(e.ParamError):
            self.query.results(policy={"expressions": expr})



# Insert record
bins = {"bin1": 0, "bin2": 1}
client.put(keyTuple, bins)

# Remove bin1
client.remove_bin(keyTuple, ['bin1'])

# Only bin2 shold remain
(keyTuple, meta, bins) = client.get(keyTuple)
print(bins)
# {'bin2': 1}



# -*- coding: utf-8 -*-

import pytest
import time
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


class TestSetQuotas(TestBaseClass):

    pytestmark = pytest.mark.skipif(
        not TestBaseClass.auth_in_use(), reason="No user specified, may be not secured cluster."
    )
    client = TestBaseClass.get_new_connection()

    def setup_method(self, method):
        """
        Setup method
        """
        usr_sys_admin_privs = [{"code": aerospike.PRIV_USER_ADMIN}, {"code": aerospike.PRIV_SYS_ADMIN}]
        try:
            self.client.admin_drop_role("usr-sys-admin-test")
            time.sleep(2)
        except Exception:
            pass

        try:
            self.client.admin_create_role("usr-sys-admin-test", usr_sys_admin_privs, write_quota=4500)
        except e.QuotasNotEnabled:
            pytest.mark.skip(reason="Got QuotasNotEnabled, skipping quota test.")
            pytest.skip()

        time.sleep(1)

    def teardown_method(self, method):
        """
        Teardown method
        """
        try:
            self.client.admin_drop_role("usr-sys-admin-test")
        except Exception:
            pass

    def test_admin_set_quota_no_parameters(self):
        """
        Set quotas with no parameters.
        """
        with pytest.raises(TypeError):
            self.client.admin_set_quotas()

    def test_admin_set_quota_no_quotas_positive(self):
        """
        Set quotas with no quotas. (will reset quotas on a role with existing quotas)
        """
        self.client.admin_set_quotas(
            role="usr-sys-admin-test",
        )
        time.sleep(1)
        roles = self.client.admin_get_role("usr-sys-admin-test")
        assert roles == {
            "privileges": [{"ns": "", "set": "", "code": 0}, {"ns": "", "set": "", "code": 1}],
            "whitelist": [],
            "read_quota": 0,
            "write_quota": 4500,
        }

    def test_admin_set_quota_one_quota_positive(self):
        """
        Set quotas with one quota.
        """
        self.client.admin_set_quotas(role="usr-sys-admin-test", read_quota=250)
        time.sleep(1)
        roles = self.client.admin_get_role("usr-sys-admin-test")
        assert roles == {
            "privileges": [{"ns": "", "set": "", "code": 0}, {"ns": "", "set": "", "code": 1}],
            "whitelist": [],
            "read_quota": 250,
            "write_quota": 4500,
        }

    def test_admin_set_quota_positive(self):
        """
        Set Quota positive
        """
        self.client.admin_set_quotas(role="usr-sys-admin-test", read_quota=250, write_quota=300)
        time.sleep(1)
        roles = self.client.admin_get_role("usr-sys-admin-test")
        assert roles == {
            "privileges": [{"ns": "", "set": "", "code": 0}, {"ns": "", "set": "", "code": 1}],
            "whitelist": [],
            "read_quota": 250,
            "write_quota": 300,
        }

    def test_admin_set_quota_positive_reset(self):
        """
        Set Quota positive
        """
        self.client.admin_set_quotas(role="usr-sys-admin-test", read_quota=0, write_quota=0)
        time.sleep(1)
        roles = self.client.admin_get_role("usr-sys-admin-test")
        assert roles == {
            "privileges": [{"ns": "", "set": "", "code": 0}, {"ns": "", "set": "", "code": 1}],
            "whitelist": [],
            "read_quota": 0,
            "write_quota": 0,
        }

    def test_admin_set_quota_positive_with_policy(self):
        """
        Set Quota positive policy
        """
        self.client.admin_set_quotas(
            role="usr-sys-admin-test", read_quota=250, write_quota=300, policy={"timeout": 1000}
        )
        time.sleep(1)
        roles = self.client.admin_get_role("usr-sys-admin-test")
        assert roles == {
            "privileges": [{"ns": "", "set": "", "code": 0}, {"ns": "", "set": "", "code": 1}],
            "whitelist": [],
            "read_quota": 250,
            "write_quota": 300,
        }

    def test_admin_set_quota_incorrect_role_name(self):
        """
        Incorrect role name
        """
        try:
            self.client.admin_set_quotas(
                role="bad-role-name", read_quota=250, write_quota=300, policy={"timeout": 1000}
            )

        except e.InvalidRole as exception:
            assert exception.code == 70
            assert exception.msg == "AEROSPIKE_INVALID_ROLE"

    def test_admin_set_quota_incorrect_role_type(self):
        """
        Incorrect role type
        """
        try:
            self.client.admin_set_quotas(role=None, read_quota=250, write_quota=300, policy={"timeout": 1000})
        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Role name should be a string."

    def test_admin_set_quota_incorrect_quota(self):
        """
        Incorrect role name
        """
        try:
            self.client.admin_set_quotas(
                role="usr-sys-admin-test", read_quota=-20, write_quota=300, policy={"timeout": 1000}
            )

        except e.InvalidRole as exception:
            assert exception.code == 70
            assert exception.msg == "AEROSPIKE_INVALID_ROLE"

    def test_admin_set_quota_incorrect_quota_type(self):
        """
        Incorrect role type
        """
        try:
            self.client.admin_set_quotas(
                role="usr-sys-admin-test", read_quota=None, write_quota=300, policy={"timeout": 1000}
            )

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Read_quota must be an integer."



# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass
from aerospike import exception as e
from aerospike_helpers.expressions import (
    And,
    BinExists,
    BinType,
    BoolBin,
    Cond,
    Def,
    DeviceSize,
    DigestMod,
    Eq,
    Exclusive,
    GE,
    GT,
    IntBin,
    IsTombstone,
    KeyBlob,
    KeyExists,
    KeyInt,
    KeyStr,
    LT,
    LastUpdateTime,
    Let,
    ListBin,
    NE,
    Not,
    Or,
    SetName,
    SinceUpdateTime,
    TTL,
    Unknown,
    Var,
    VoidTime,
)
from aerospike_helpers.operations import expression_operations as expressions
from aerospike_helpers.operations import operations

import aerospike


# Constants
_NUM_RECORDS = 8


def verify_multiple_expression_avenues(client, test_ns, test_set, expr, op_bin, expected):
    keys = [(test_ns, test_set, i) for i in range(_NUM_RECORDS + 1)]

    # operate
    ops = [operations.read(op_bin)]
    res = []
    for key in keys:
        try:
            res.append(client.operate(key, ops, policy={"expressions": expr})[2])
        except e.FilteredOut:
            pass

    assert len(res) == expected

    # operate ordered
    ops = [operations.read(op_bin)]
    res = []
    for key in keys:
        try:
            res.append(client.operate_ordered(key, ops, policy={"expressions": expr})[2])
        except e.FilteredOut:
            pass

    # batch get
    res = [rec for rec in client.get_many(keys, policy={"expressions": expr}) if rec[2]]

    assert len(res) == expected

    # scan results
    scan_obj = client.scan(test_ns, test_set)
    records = scan_obj.results({"expressions": expr})
    assert len(records) == expected

    # other scan methods
    # execute_background tested test_scan_execute_background.py
    # foreach tested test_scan.py

    # query results
    query_obj = client.query(test_ns, test_set)
    records = query_obj.results({"expressions": expr})
    assert len(records) == expected

    # other query methods
    # execute background tested in test_query_execute_background.py
    # foreach tested in test_query.py


class TestExpressions(TestBaseClass):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        self.test_ns = "test"
        self.test_set = "demo"

        for i in range(_NUM_RECORDS):
            key = ("test", "demo", i)
            rec = {
                "name": "name%s" % (str(i)),
                "t": True,
                "f": False,
                "age": i,
                "balance": i * 10,
                "key": i,
                "alt_name": "name%s" % (str(i)),
                "list_bin": [
                    None,
                    i,
                    "string_test" + str(i),
                    [26, 27, 28, i],
                    {31: 31, 32: 32, 33: 33, i: i},
                    bytearray("bytearray_test" + str(i), "utf8"),
                    ("bytes_test" + str(i)).encode("utf8"),
                    i % 2 == 1,
                    aerospike.null,
                    float(i),
                ],
                "ilist_bin": [
                    1,
                    2,
                    6,
                ],
                "imap_bin": {
                    1: 1,
                    2: 2,
                    3: 6,
                },
            }
            self.as_connection.put(key, rec)

        self.as_connection.put(
            ("test", "demo", _NUM_RECORDS), {"extra": "record"}, policy={"key": aerospike.POLICY_KEY_SEND}
        )
        self.as_connection.put(
            ("test", "demo", "k_string"), {"test": "data"}, policy={"key": aerospike.POLICY_KEY_SEND}
        )
        self.as_connection.put(
            ("test", "demo", bytearray("b_string", "utf-8")),
            {"test": "b_data"},
            policy={"key": aerospike.POLICY_KEY_SEND},
        )

        def teardown():
            for i in range(_NUM_RECORDS):
                key = ("test", "demo", i)
                as_connection.remove(key)

            as_connection.remove(("test", "demo", _NUM_RECORDS))
            as_connection.remove(("test", "demo", "k_string"))
            as_connection.remove(("test", "demo", bytearray("b_string", "utf-8")))

        request.addfinalizer(teardown)

    @pytest.mark.xfail(reason="Will fail on storage engine device.")
    def test_device_size_pos(self):
        expr = Eq(DeviceSize(), 0)
        record = self.as_connection.get(("test", "demo", _NUM_RECORDS), policy={"expressions": expr.compile()})
        assert record[2]["extra"] == "record"

    def test_TTL_pos(self):
        expr = NE(TTL(), 0)
        record = self.as_connection.get(("test", "demo", _NUM_RECORDS), policy={"expressions": expr.compile()})
        assert record[2]["extra"] == "record"

    def test_void_time_pos(self):
        expr = NE(VoidTime(), 0)
        record = self.as_connection.get(("test", "demo", _NUM_RECORDS), policy={"expressions": expr.compile()})
        assert record[2]["extra"] == "record"

    def test_remove_with_expressions_neg(self):
        self.as_connection.put(("test", "demo", 25), {"test": "test_data"})

        expr = Eq(KeyInt(), 26)
        with pytest.raises(e.FilteredOut):
            self.as_connection.remove(("test", "demo", 25), policy={"expressions": expr.compile()})

    def test_scan_with_results_method_and_expressions(self):
        ns = "test"
        st = "demo"

        expr = Eq(KeyInt(), _NUM_RECORDS)
        record = self.as_connection.get(("test", "demo", _NUM_RECORDS), policy={"expressions": expr.compile()})
        assert record[2]["extra"] == "record"

        expr = Eq(KeyStr(), "k_string")
        record = self.as_connection.get(("test", "demo", "k_string"), policy={"expressions": expr.compile()})
        assert record[2]["test"] == "data"

        expr = Eq(KeyBlob(), bytearray("b_string", "utf-8"))
        record = self.as_connection.get(
            ("test", "demo", bytearray("b_string", "utf-8")), policy={"expressions": expr.compile()}
        )
        assert record[2]["test"] == "b_data"

        expr = KeyExists()
        record = self.as_connection.get(
            ("test", "demo", bytearray("b_string", "utf-8")), policy={"expressions": expr.compile()}
        )
        assert record[2]["test"] == "b_data"

        expr = And(
            BinExists("age"),
            Eq(SetName(), "demo"),
            NE(LastUpdateTime(), 0),
            NE(SinceUpdateTime(), 0),
            Not(IsTombstone()),
            Eq(DigestMod(2), 0),
        )

        scan_obj = self.as_connection.scan(ns, st)
        records = scan_obj.results({"expressions": expr.compile()})
        assert 5 == len(records)

    @pytest.mark.parametrize(
        "bin, expected_bin_type",
        [
            ("ilist_bin", aerospike.AS_BYTES_LIST),
            ("age", aerospike.AS_BYTES_INTEGER),
            ("imap_bin", aerospike.AS_BYTES_MAP),
        ],
    )
    def test_bin_type_pos(self, bin, expected_bin_type):
        """
        Invoke BinType() on various kinds of bins.
        """
        expr = Eq(BinType(bin), expected_bin_type).compile()
        verify_multiple_expression_avenues(self.as_connection, self.test_ns, self.test_set, expr, bin, _NUM_RECORDS)

    def test_nested_logic_pos(self):
        """
        Test nested logical operators expression.
        """

        expr = Or(
            Or(Eq(ListBin("ilist_bin"), [1, 2, 7]), Eq(ListBin("ilist_bin"), [1, 2, 6])),
            And(LT(IntBin("age"), 22), GT(IntBin("age"), -1)),
            And(
                Or(Eq(ListBin("ilist_bin"), [1, 2, 7]), Eq(ListBin("ilist_bin"), [1, 2, 6])),
                And(LT(IntBin("age"), 22), GT(IntBin("age"), -1)),
            ),
        ).compile()
        verify_multiple_expression_avenues(
            self.as_connection, self.test_ns, self.test_set, expr, "ilist_bin", _NUM_RECORDS
        )

    def test_bool_bin_true(self):
        if self.server_version < [5, 6]:
            pytest.mark.xfail(reason="Servers older than 5.6 do not support 6.0.0 expressions")
            pytest.xfail()

        config = TestBaseClass.get_connection_config()
        config["send_bool_as"] = aerospike.AS_BOOL
        test_client = aerospike.client(config).connect(config["user"], config["password"])

        expr = BoolBin("t")
        ops = [operations.write("t", True), expressions.expression_read("test", expr.compile())]
        _, _, res = test_client.operate(("test", "demo", _NUM_RECORDS - 1), ops)
        test_client.close()
        assert res["test"]

    def test_bool_bin_false(self):
        if self.server_version < [5, 6]:
            pytest.mark.xfail(reason="Servers older than 5.6 do not support 6.0.0 expressions")
            pytest.xfail()

        config = TestBaseClass.get_connection_config()
        config["send_bool_as"] = aerospike.AS_BOOL
        test_client = aerospike.client(config).connect(config["user"], config["password"])

        expr = Not(BoolBin("t"))
        ops = [operations.write("t", True), expressions.expression_read("test", expr.compile())]
        _, _, res = test_client.operate(("test", "demo", _NUM_RECORDS - 1), ops)
        test_client.close()
        assert not res["test"]

    def test_exclusive_pos(self):
        if self.server_version < [5, 6]:
            pytest.mark.xfail(reason="Servers older than 5.6 do not support 6.0.0 expressions")
            pytest.xfail()

        expr = Exclusive(GT(IntBin("age"), _NUM_RECORDS // 2), GT(IntBin("age"), _NUM_RECORDS - 1))
        record = self.as_connection.get(("test", "demo", _NUM_RECORDS - 2), policy={"expressions": expr.compile()})
        assert record[2]["age"] == _NUM_RECORDS - 2

    def test_exclusive_neg(self):
        if self.server_version < [5, 6]:
            pytest.mark.xfail(reason="Servers older than 5.6 do not support 6.0.0 expressions")
            pytest.xfail()

        expr = Exclusive(GT(IntBin("age"), _NUM_RECORDS // 2), GT(IntBin("age"), _NUM_RECORDS // 2))
        with pytest.raises(e.FilteredOut):
            self.as_connection.get(("test", "demo", _NUM_RECORDS - 2), policy={"expressions": expr.compile()})

    def test_let_def_var_pos(self):
        if self.server_version < [5, 6]:
            pytest.mark.xfail(reason="Servers older than 5.6 do not support 6.0.0 expressions")
            pytest.xfail()

        expr = Let(Def("a", IntBin("age")), Cond(LT(Var("a"), 50), True, Unknown()))
        record = self.as_connection.get(("test", "demo", _NUM_RECORDS - 1), policy={"expressions": expr.compile()})
        assert record[2]["age"] == _NUM_RECORDS - 1

    def test_let_def_var_neg(self):
        if self.server_version < [5, 6]:
            pytest.mark.xfail(reason="Servers older than 5.6 do not support 6.0.0 expressions")
            pytest.xfail()

        expr = Let(Def("a", IntBin("age")), Cond(LT(Var("a"), 0), True, Unknown()))
        with pytest.raises(e.FilteredOut):
            self.as_connection.get(("test", "demo", _NUM_RECORDS - 1), policy={"expressions": expr.compile()})

    def test_cond_pos(self):
        if self.server_version < [5, 6]:
            pytest.mark.xfail(reason="Servers older than 5.6 do not support 6.0.0 expressions")
            pytest.xfail()

        expr = Cond(GE(IntBin("age"), 2), True, Unknown())
        record = self.as_connection.get(("test", "demo", _NUM_RECORDS - 1), policy={"expressions": expr.compile()})
        assert record[2]["age"] == _NUM_RECORDS - 1

    def test_cond_neg(self):
        if self.server_version < [5, 6]:
            pytest.mark.xfail(reason="Servers older than 5.6 do not support 6.0.0 expressions")
            pytest.xfail()

        expr = Cond(GT(IntBin("age"), _NUM_RECORDS), True, Unknown())
        with pytest.raises(e.FilteredOut):
            self.as_connection.get(("test", "demo", _NUM_RECORDS - 1), policy={"expressions": expr.compile()})

    def test_bintype_as_bool(self):
        if self.server_version < [5, 6]:
            pytest.mark.xfail(reason="Servers older than 5.6 do not support 6.0.0 expressions")
            pytest.xfail()

        # Configure client to encode and send booleans as the server boolean type
        config = TestBaseClass.get_connection_config()
        config["send_bool_as"] = aerospike.AS_BOOL
        test_client = aerospike.client(config).connect(config["user"], config["password"])

        # Override record 0's "t" bin to be a server bool instead of a python bool
        key = ("test", "demo", 0)
        bins = {"t": True}
        test_client.put(key, bins)

        # Check that record 0 has a server boolean bin named "t"
        expr = Eq(BinType("t"), aerospike.AS_BYTES_BOOL).compile()
        records = test_client.get_many([key], {"expressions": expr})

        # bins would be None if the record was filtered out by the expression
        record = records[0]
        bins = record[2]
        assert bins

        test_client.close()



import aerospike

config = {'hosts': [('127.0.0.1', 3000)],
            'lua': {'system_path':'/usr/local/aerospike/lua/',
                    'user_path':'./'}}
client = aerospike.client(config).connect()
client.udf_put("example.lua")

# Remove index if it already exists
from aerospike import exception as ex
try:
    client.index_remove("test", "ageIndex")
except ex.IndexNotFound:
    pass

bins = [
    {"name": "Jeff", "age": 20},
    {"name": "Derek", "age": 24},
    {"name": "Derek", "age": 21},
    {"name": "Derek", "age": 29},
    {"name": "Jeff", "age": 29},
]
keys = [("test", "users", f"user{i}") for i in range(len(bins))]
for key, recordBins in zip(keys, bins):
    client.put(key, recordBins)

client.index_integer_create("test", "users", "age", "ageIndex")

query = client.query('test', 'users')
query.apply('example', 'group_count', ['name', 'age', 21])
names = query.results()

# we expect a dict (map) whose keys are names, each with a count value
print(names)
# One of the Jeffs is excluded because he is under 21
# [{'Derek': 3, 'Jeff': 1}]

# Cleanup
client.index_remove("test", "ageIndex")
client.batch_remove(keys)
client.close()



# Insert 4 records with these keys
keyTuples = [
    ('test', 'demo', 1),
    ('test', 'demo', 2),
    ('test', 'demo', 3),
    ('test', 'demo', 4)
]
# Only records 1, 2, 4 have a bin called bin2
client.put(keyTuples[0], {'bin1': 20, 'bin2': 40})
client.put(keyTuples[1], {'bin1': 11, 'bin2': 50})
client.put(keyTuples[2], {'bin1': 50,             'bin3': 20})
client.put(keyTuples[3], {'bin1': 87, 'bin2': 76, 'bin3': 40})

# Get all 4 records and filter out every bin except bin2
records = client.select_many(keyTuples, ['bin2'])
for record in records:
    print(record)

# (('test', 'demo', 1, bytearray(...)), {'ttl': 2592000, 'gen': 1}, {'bin2': 40})
# (('test', 'demo', 2, bytearray(...)), {'ttl': 2592000, 'gen': 1}, {'bin2': 50})
# (('test', 'demo', 3, bytearray(...)), {'ttl': 2592000, 'gen': 1}, {})
# (('test', 'demo', 4, bytearray(...)), {'ttl': 2592000, 'gen': 1}, {'bin2': 76})


# -*- coding: utf-8 -*-

import pytest
from contextlib import contextmanager

try:
    import cPickle as pickle
except Exception:
    import pickle
from . import test_data

# from collections import OrderedDict
from .test_base_class import TestBaseClass

import aerospike
from aerospike import exception as e


@contextmanager
def open_as_connection(config):
    """
    Context manager to let us open aerospike connections with
    specified config
    """
    as_connection = TestBaseClass.get_new_connection(config)

    # Connection is setup, so yield it
    yield as_connection

    # close the connection
    as_connection.close()


class SomeClass(object):
    pass


@pytest.mark.usefixtures("as_connection")
# adds cls.connection_config to this class
@pytest.mark.usefixtures("connection_config")
class TestGetPut:
    @pytest.mark.parametrize("_input, _expected", test_data.pos_data)
    def test_pos_get_put_with_key(self, _input, _expected, put_data):
        """
        Invoke get() with a key and not policy's dict.
        """
        put_data(self.as_connection, _input, _expected)
        _, _, bins = self.as_connection.get(_input)
        assert bins == _expected

    def test_pos_get_with_data_missing(self):
        """
        Invoke get with a record which does not exist.
        """
        key = ("test", "demo", "non-existent-key-that-does-not-exist")
        with pytest.raises(e.RecordNotFound):
            self.as_connection.get(key)

    @pytest.mark.skip(reason="byte key not currently handled")
    def test_get_information_using_bytes_key(self):
        record = {"bytes": "are_cool"}
        key = ("test", "demo", b"\x83X\xb5\xde")
        self.as_connection.put(key, record)
        _, _, rec = self.as_connection.get(key)

        assert rec == record
        self.as_connection.remove(key)

    def test_pos_get_initkey_with_digest(self, put_data):
        """
        Invoke get() for a record having bytestring data.
        """
        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"))
        rec = {"name": "john"}

        policy = {"key": aerospike.POLICY_KEY_DIGEST}

        put_data(self.as_connection, key, rec, policy)

        key, _, bins = self.as_connection.get(key, policy)

        assert bins == {"name": "john"}
        assert key == ("test", "demo", None, bytearray(b"asd;as[d'as;djk;uyfl"))

    @pytest.mark.parametrize(
        "_input, _record, _policy, _expected",
        [
            (("test", "demo", 3), {"name": "name%s" % (str(3)), "age": 3}, aerospike.POLICY_KEY_DIGEST, None),
            (("test", "demo", 3), {"name": "name%s" % (str(3)), "age": 3}, aerospike.POLICY_KEY_SEND, 3),
        ],
    )
    def test_pos_get_with_policy_key_digest(self, _input, _record, _policy, _expected, put_data):
        """
        Invoke get() for a record with POLICY_KEY_DIGEST
        """
        put_data(self.as_connection, _input, _record)

        (key, _, _) = self.as_connection.get(_input, policy={"key": _policy})
        assert key[2] == _expected

    def test_pos_get_initkey_with_policy_send(self, put_data):
        """
        Invoke get() for a record having string data.
        """

        key = ("test", "demo", 1)

        rec = {"name": "john", "age": 1}

        policy = {"key": aerospike.POLICY_KEY_SEND}

        put_data(self.as_connection, key, rec, policy)

        key, _, bins = self.as_connection.get(key, policy)

        assert bins == {"name": "john", "age": 1}
        assert key == (
            "test",
            "demo",
            1,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_pos_get_initkey_with_client_policy_send(self, put_data):
        """
        Invoke get() for a record having string data.
        """

        key = ("test", "demo", 1)

        rec = {"name": "john", "age": 1}

        put_data(self.as_connection, key, rec)

        key, _, bins = self.as_connection.get(key)

        assert bins == {"name": "john", "age": 1}
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

        config = self.connection_config.copy()
        config["policies"] = {"key": aerospike.POLICY_KEY_SEND}

        with open_as_connection(config) as client:
            assert client is not None
            assert client.is_connected()

            key = ("test", "demo", 2)
            rec = {"name": "john", "age": 2}

            put_data(client, key, rec)
            key, _, bins = client.get(key)
            assert bins == {"name": "john", "age": 2}
            assert key == ("test", "demo", 2, bytearray(b"\xaejQ_7\xdeJ\xda\xccD\x96\xe2\xda\x1f\xea\x84\x8c:\x92p"))

    # Negative get tests
    def test_neg_get_with_no_parameter(self):
        """
        Invoke get() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.get()

        assert "argument 'key' (pos 1)" in str(typeError.value)

    def test_neg_get_with_extra_parameter_in_key(self, put_data):
        """
        Invoke get() with extra parameter
        """
        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"), 1)
        rec = {"name": "john"}

        policy = {"key": aerospike.POLICY_KEY_DIGEST}
        with pytest.raises(e.ParamError):
            put_data(self.as_connection, key, rec, policy)

    def test_neg_get_with_key_digest(self):
        """
        Invoke get() with a key digest.
        """
        key = ("test", "demo", 1)
        try:
            key, _ = self.as_connection.exists(key)
            key, _, _ = self.as_connection.get((key[0], key[1], None, key[2]))
        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "digest is invalid. expected a bytearray"
        except e.RecordNotFound as exception:
            assert exception.code == 2

    @pytest.mark.parametrize("key, ex_code, ex_msg", test_data.key_neg)
    def test_neg_get_with_none(self, key, ex_code, ex_msg):
        """
        Invoke get() with None namespace/key in key tuple.
        """
        with pytest.raises(e.ParamError):
            self.as_connection.get(key)

    def test_neg_get_with_invalid_record(self):
        """
        Get a record which does not exist
        """
        key = ("test", "demo", "A PRIMARY KEY WHICH DOES NOT EXIST")
        with pytest.raises(e.RecordNotFound):
            self.as_connection.get(key)

    def test_neg_get_with_non_existent_namespace(self):
        """
        Invoke get() for non-existent namespace.
        """
        key = ("namespace", "demo", 1)

        with pytest.raises(e.ClientError):
            key, _, _ = self.as_connection.get(key)

    @pytest.mark.parametrize(
        "_input, _expected",
        [
            (("test", "demo", 1), {"name": "john", "age": 1}),
        ],
    )
    def test_neg_get_remove_key_and_check_get(self, _input, _expected, put_data):
        """
        Invoke get() for a record having string data.
        """
        put_data(self.as_connection, _input, _expected)
        self.as_connection.remove(_input)
        try:
            _, _, bins = self.as_connection.get(_input)
            assert bins is None
        except e.RecordNotFound as exception:
            assert exception.code == 2

    def test_neg_get_with_only_key_no_connection(self):
        """
        Invoke get() with a key and not policy's dict no connection
        """
        key = ("test", "demo", 1)
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        with pytest.raises(e.ClusterError):
            key, _, _ = client1.get(key)

    # Put Tests
    def test_pos_put_with_policy_exists_create_or_replace(self):
        """
        Invoke put() for a record with create or replace policy positive
        If record exists and will be replaced
        """
        key = ("test", "demo", 1)

        rec = {"name": "Smith"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "exists": aerospike.POLICY_EXISTS_CREATE_OR_REPLACE,
            "gen": aerospike.POLICY_GEN_IGNORE,
            "key": aerospike.POLICY_KEY_SEND,
        }
        assert 0 == self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)
        assert rec == bins

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "exists": aerospike.POLICY_EXISTS_CREATE_OR_REPLACE,
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }
        assert 0 == self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)
        assert rec == bins
        self.as_connection.remove(key)

    def test_pos_put_with_policy_exists_ignore_create(self):
        """
        Invoke put() for a record with ignore.
        """
        key = ("test", "demo", 1)

        rec = {"name": "Smith"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "exists": aerospike.POLICY_EXISTS_IGNORE,
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }
        assert 0 == self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)
        assert rec == bins
        self.as_connection.remove(key)

    def test_pos_put_with_policy_exists_ignore_update(self):
        """
        Invoke put() for a record with ignore.
        Ignore for an existing records should update without any error
        """
        key = ("test", "demo", 1)

        rec = {"name": "Smith"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "exists": aerospike.POLICY_EXISTS_IGNORE,
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }
        assert 0 == self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)
        assert rec == bins

        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "exists": aerospike.POLICY_EXISTS_IGNORE,
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }
        assert 0 == self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)
        assert rec == bins
        self.as_connection.remove(key)

    def test_pos_put_with_policy_replace(self):
        """
        Invoke put() for a record with replace positive.
        """
        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }
        assert 0 == self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        assert rec == bins

        rec = {"name": "Smith"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {"exists": aerospike.POLICY_EXISTS_REPLACE}
        assert 0 == self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        assert rec == bins
        self.as_connection.remove(key)

    def test_pos_put_with_policy_exists_update_positive(self):
        """
        Invoke put() for a record with all policies.
        """
        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }
        assert 0 == self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        assert rec == bins

        rec = {"name": "Smith"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {"exists": aerospike.POLICY_EXISTS_UPDATE}
        assert 0 == self.as_connection.put(key, rec, meta, policy)
        (key, meta, bins) = self.as_connection.get(key)

        assert rec == bins
        self.as_connection.remove(key)

    def test_pos_put_with_policy_gen_GT(self):
        """
        Invoke put() for a record with generation as GT positive
        """
        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {}
        assert 0 == self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        assert rec == bins
        gen = meta["gen"]
        assert gen == 1
        rec = {"name": "Smith"}
        policy = {"gen": aerospike.POLICY_GEN_GT}
        meta = {"gen": gen + 5}

        self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)
        assert rec == bins
        self.as_connection.remove(key)

    def test_pos_put_with_policy_gen_ignore(self):
        """
        Invoke put() for a record with generation as gen_ignore
        """
        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {}
        assert 0 == self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        assert rec == bins

        gen = meta["gen"]
        rec = {"name": "Smith"}
        policy = {"gen": aerospike.POLICY_GEN_IGNORE}
        meta = {"gen": gen}

        assert 0 == self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)
        assert rec == bins
        self.as_connection.remove(key)

    @pytest.mark.parametrize(
        "key, record, meta",
        [
            (("test", "demo", 1), {"name": "john"}, {"gen": True, "ttl": 25000}),
            (("test", "demo", 1), {"name": "john"}, {"gen": 3, "ttl": True}),
            (("test", "demo", 1), {"name": "john"}, {"gen": True, "ttl": True}),
            (
                ("test", "demo", 1),
                {"name": "john"},
                {"gen": True, "ttl": aerospike.TTL_NAMESPACE_DEFAULT},
            ),
            (
                ("test", "demo", 1),
                {"name": "john"},
                {"gen": True, "ttl": aerospike.TTL_NEVER_EXPIRE},
            ),
        ],
    )
    def test_pos_put_with_metadata_bool(self, key, record, meta, put_data):
        """
        Invoke put() for a record with generation as boolean.
        """
        put_data(self.as_connection, key, record, meta)

        (key, meta, bins) = self.as_connection.get(key)
        assert bins == record
        assert meta["gen"] is not None
        assert meta["ttl"] is not None

    def test_pos_put_user_serializer_no_deserializer(self):
        """
        Invoke put() for float data record with user serializer is
        registered, but deserializer is not registered.
        """
        key = ("test", "demo", "put_user_serializer")

        rec = {"pi": 3.14}

        def serialize_function(val):
            return pickle.dumps(val)

        aerospike.set_serializer(serialize_function)

        res = self.as_connection.put(key, rec, {}, {}, aerospike.SERIALIZER_USER)

        assert res == 0

        _, _, bins = self.as_connection.get(key)

        if self.skip_old_server is False:
            assert bins == {"pi": 3.14}
        else:
            assert bins == {"pi": bytearray(b"F3.1400000000000001\n.")}

        self.as_connection.remove(key)

    # put negative
    def test_neg_put_with_no_parameters(self):
        """
        Invoke put() without any parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.put()

        assert "argument 'key' (pos 1)" in str(typeError.value)

    def test_neg_put_without_record(self):
        """
        Invoke put() without any record data.
        """
        key = ("test", "demo", 1)

        with pytest.raises(TypeError) as typeError:
            self.as_connection.put(key)

        assert "argument 'bins' (pos 2)" in str(typeError.value)

    @pytest.mark.parametrize("key, ex_code, ex_msg", test_data.key_neg)
    def test_neg_put_with_none(self, key, ex_code, ex_msg, record={}):
        """
        Invoke put() with invalid data
        """

        with pytest.raises(e.ParamError):
            self.as_connection.put(key, record)

    @pytest.mark.parametrize(
        "key, ex_code, ex_msg, record",
        [
            (("test", "demo", None), -2, "either key or digest is required", {"name": "John"}),
        ],
    )
    def test_neg_put_with_invalid_record(self, key, ex_code, ex_msg, record):
        """
        Invoke put() with invalid data
        """

        with pytest.raises(e.ParamError):
            self.as_connection.put(key, record)

    @pytest.mark.parametrize(
        "key, record, exception_code",
        [
            # Non-existing NS & Set
            (("demo", "test", 1), {"a": ["!@#!#$%#", bytearray("ASD@#$AR#$@#ERQ#", "utf-8")]}, 20),
            # Non-existing Namespace
            (("test1", "demo", 1), {"i": "asdadasd"}, 20),
        ],
    )
    def test_neg_put_with_wrong_ns_and_set(self, key, record, exception_code):
        """
        Invoke put() with non-existent data
        """
        with pytest.raises(e.ClientError):
            self.as_connection.put(key, record)

    def test_neg_put_with_policy_gen_EQ_less(self):
        """
        Invoke put() for a record with generation as EQ less
        """
        key = ("test", "demo", "policy_gen_EQ_key")

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {}
        assert 0 == self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        assert {"name": "John"} == bins
        rec = {"name": "Smith"}
        policy = {"gen": aerospike.POLICY_GEN_EQ}
        meta = {"gen": 10}

        with pytest.raises(e.RecordGenerationError):
            self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)
        assert {"name": "John"} == bins

        self.as_connection.remove(key)

    def test_neg_put_with_policy_gen_EQ_more(self):
        """
        Invoke put() for a record with generation as EQ less
        """
        key = ("test", "demo", "policy_gen_EQ_more_key")

        rec = {"name": "John"}
        meta = {"gen": 10, "ttl": 25000}
        assert 0 == self.as_connection.put(key, rec, meta)

        (key, meta, bins) = self.as_connection.get(key)

        assert {"name": "John"} == bins
        rec = {"name": "Smith"}
        policy = {"gen": aerospike.POLICY_GEN_EQ}
        meta = {"gen": 4}

        with pytest.raises(e.RecordGenerationError):
            self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)
        assert {"name": "John"} == bins

        self.as_connection.remove(key)

    def test_neg_put_with_policy_exists_create(self):
        """
        Invoke put() for a record with all policies.
        """
        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }
        assert 0 == self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        assert {"name": "John"} == bins

        rec = {"name": "Smith"}
        policy = {"exists": aerospike.POLICY_EXISTS_CREATE}
        meta = {"gen": 2}

        with pytest.raises(e.RecordExistsError):
            self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)
        assert {"name": "John"} == bins

        self.as_connection.remove(key)

    def test_neg_put_with_policy_exists_replace_negative(self):
        """
        Invoke put() for a record with replace policy negative.
        """
        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "exists": aerospike.POLICY_EXISTS_REPLACE,
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }
        with pytest.raises(e.RecordNotFound):
            assert 0 == self.as_connection.put(key, rec, meta, policy)

    def test_neg_put_with_policy_replace(self):
        """
        Invoke put() for a record with replace positive.
        """
        key = ("test", "demo", 1)

        rec = {"name": "Smith"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {"exists": aerospike.POLICY_EXISTS_REPLACE}

        with pytest.raises(e.RecordNotFound):
            self.as_connection.put(key, rec, meta, policy)

    def test_neg_put_with_policy_exists_update_negative(self):
        """
        Invoke put() for a record with update policy negative.
        """
        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "exists": aerospike.POLICY_EXISTS_UPDATE,
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }
        try:
            assert 0 == self.as_connection.put(key, rec, meta, policy)

        except e.RecordNotFound as exception:
            assert exception.code == 2
            assert exception.msg == "AEROSPIKE_ERR_RECORD_NOT_FOUND"

    def test_neg_put_with_policy_gen_GT_lesser(self):
        """
        Invoke put() for a record with generation as GT lesser
        """
        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        assert 0 == self.as_connection.put(key, rec, meta)

        (key, meta, bins) = self.as_connection.get(key)

        assert {"name": "John"} == bins
        gen = meta["gen"]
        rec = {"name": "Smith"}
        policy = {"gen": aerospike.POLICY_GEN_GT}
        meta = {"gen": gen}

        try:
            self.as_connection.put(key, rec, meta, policy)

        except e.RecordGenerationError as exception:
            assert exception.code == 3
            assert exception.msg == "AEROSPIKE_ERR_RECORD_GENERATION"

        (key, meta, bins) = self.as_connection.get(key)
        assert {"name": "John"} == bins
        self.as_connection.remove(key)

    def test_neg_put_with_string_record_without_connection(self):
        """
        Invoke put() for a record with string data without connection
        """
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        key = ("test", "demo", 1)

        bins = {"name": "John"}

        try:
            client1.put(key, bins)
        except e.ClusterError as exception:
            assert exception.code == 11

    @pytest.mark.parametrize(
        "key, record, meta, policy, ex_code, ex_msg",
        [
            (
                ("test", "demo", 1),
                {"name": "john"},
                {"gen": "wrong", "ttl": 25000},
                None,
                -2,
                "Generation should be an int or long",
            ),
            (
                ("test", "demo", 1),
                {"name": "john"},
                {"gen": 3, "ttl": "25000"},
                None,
                -2,
                "TTL should be an int or long",
            ),
            (
                ("test", "demo", 1),
                {"name": "john"},
                {"gen": 3, "ttl": 25000},
                {"total_timeout": 0.5},
                -2,
                "timeout is invalid",
            ),
            (
                ("test", "demo", 1),
                {"name": "john"},  # Policy as string
                {"gen": 3, "ttl": 25000},
                "Policy",
                -2,
                "policy must be a dict",
            ),
            (
                ("test", "demo", 1),
                {"i": 13},  # Meta as string
                "OK",
                None,
                -2,
                "meta must be a dict",
            ),
            (
                ("test", "demo", 1),
                {"i": 13},  # Meta as string
                1234,
                None,
                -2,
                "meta must be a dict",
            ),
        ],
    )
    def test_neg_put_with_invalid_metadata(self, key, record, meta, policy, ex_code, ex_msg, put_data):
        """
        Invoke put() for a record with generation as string
        """
        with pytest.raises(e.ParamError):
            put_data(self.as_connection, key, record, meta, policy)
        #     # self.as_connection.remove(key)
        # except e.ParamError as exception:
        #     assert exception.code == ex_code
        #     assert exception.msg == ex_msg

    # put edge cases
    def test_edge_put_record_with_bin_name_exceeding_max_limit(self):
        """
        Invoke put() with bin name exceeding the max limit of bin name.
        """
        key = ("test", "demo", "put_rec")
        put_record = {"a" * 50: "unimportant"}

        with pytest.raises(e.BinNameError):
            self.as_connection.put(key, put_record)

    def test_edge_put_with_integer_greater_than_maxisze(self):
        """
        Invoke put() for a record with integer greater than max size
        """
        key = ("test", "demo", 1)

        bins = {"no": 111111111111111111111111111111111111111111111}

        try:
            assert 0 == self.as_connection.put(key, bins)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "integer value exceeds sys.maxsize"
        except SystemError:
            pass

    def test_edge_put_with_key_as_an_integer_greater_than_maxsize(self):
        """
        Invoke put() for a record with integer greater than max size
        """
        key = ("test", "demo", 1111111111111111111111111111111111)

        bins = {"no": 11}

        with pytest.raises(e.ParamError):
            assert 0 == self.as_connection.put(key, bins)

    def test_unhashable_type_with_put_get(self):
        """
        Invoke put() with bytes which used to be read as bytearray off server
        causing an unhashable key error on read. Now server byte blobs are read as bytes.
        """
        key = ("test", "demo", 123)

        bins = {"bin1": b"abc", "bin2": {b"abc": 1}}

        try:
            assert 0 == self.as_connection.put(key, bins)
            res = self.as_connection.get(key)
            bin2 = res[2]["bin2"]
            assert bin2[b"abc"]
        except Exception as e:
            raise (e)
        finally:
            self.as_connection.remove(key)

    def test_null_bin_with_put_get(self):
        """
        Invoke put() with null bin and verify record is deleted.
        """
        key = ("test", "demo", 123)

        bin = {"bin1": "abc"}
        null_bin = {"bin1": aerospike.null()}

        assert 0 == self.as_connection.put(key, bin)
        res = self.as_connection.get(key)
        bin1 = res[2]["bin1"]
        assert bin1 == "abc"

        assert 0 == self.as_connection.put(key, null_bin)
        with pytest.raises(e.RecordNotFound):
            res = self.as_connection.get(key)



# -*- coding: utf-8 -*-

import pytest
import time
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


@pytest.mark.usefixtures("connection_config")
class TestCreateRole(object):

    config = TestBaseClass.get_connection_config()

    pytestmark = pytest.mark.skipif(
        not TestBaseClass.auth_in_use(), reason="No user specified, may be not secured cluster."
    )

    def setup_method(self, method):
        """
        Setup method
        """
        config = self.config
        self.client = aerospike.client(config).connect(config["user"], config["password"])
        try:
            self.client.admin_drop_user("testcreaterole")
            time.sleep(2)
        except Exception:
            pass  # do nothing, EAFP

        self.delete_users = []

    def teardown_method(self, method):
        """
        Teardown method
        """

        policy = {}

        for user in self.delete_users:
            try:
                self.client.admin_drop_user(user, policy)
            except Exception:
                pass

        self.client.close()

    def test_create_role_without_any_parameters(self):

        with pytest.raises(TypeError) as typeError:
            self.client.admin_create_role()

        assert "argument 'role' (pos 1)" in str(typeError.value)

    def test_create_role_positive_with_policy(self):
        """
        Create role positive
        """
        try:
            self.client.admin_get_role("usr-sys-admin-test")
            # role exists, clear it out.
            self.client.admin_drop_role("usr-sys-admin-test")
            time.sleep(2)
        except e.InvalidRole:
            pass  # we are good, no such role exists

        self.client.admin_create_role(
            "usr-sys-admin-test", [{"code": aerospike.PRIV_READ, "ns": "test", "set": "demo"}], {"timeout": 1000}
        )
        time.sleep(1)
        roles = self.client.admin_get_role("usr-sys-admin-test")
        assert roles == {
            "privileges": [{"code": 10, "ns": "test", "set": "demo"}],
            "whitelist": [],
            "read_quota": 0,
            "write_quota": 0,
        }

        try:
            status = self.client.admin_create_user("testcreaterole", "createrole", ["usr-sys-admin-test"])
        except e.QuotasNotEnabled:
            pytest.mark.skip(reason="Got QuotasNotEnabled, skipping quota test.")
            pytest.skip()

        assert status == 0
        time.sleep(1)
        roles = self.client.admin_query_user("testcreaterole")

        assert roles == ["usr-sys-admin-test"]

        self.client.admin_drop_user("testcreaterole")

    @pytest.mark.parametrize(
        "priv_name, privs",
        [
            ("PRIV_USER_ADMIN", [{"code": aerospike.PRIV_USER_ADMIN, "ns": "", "set": ""}]),
            ("PRIV_SYS_ADMIN", [{"code": aerospike.PRIV_SYS_ADMIN, "ns": "", "set": ""}]),
            ("PRIV_DATA_ADMIN", [{"code": aerospike.PRIV_DATA_ADMIN, "ns": "", "set": ""}]),
            ("PRIV_READ", [{"code": aerospike.PRIV_READ, "ns": "test", "set": "demo"}]),
            ("PRIV_WRITE", [{"code": aerospike.PRIV_WRITE, "ns": "test", "set": "demo"}]),
            ("PRIV_READ_WRITE", [{"code": aerospike.PRIV_READ_WRITE, "ns": "test", "set": "demo"}]),
            ("PRIV_READ_WRITE_UDF", [{"code": aerospike.PRIV_READ_WRITE_UDF, "ns": "test", "set": "demo"}]),
            ("PRIV_TRUNCATE", [{"code": aerospike.PRIV_TRUNCATE, "ns": "test", "set": "demo"}]),
            ("PRIV_UDF_ADMIN", [{"code": aerospike.PRIV_UDF_ADMIN, "ns": "", "set": ""}]),
            ("PRIV_SINDEX_ADMIN", [{"code": aerospike.PRIV_SINDEX_ADMIN, "ns": "", "set": ""}]),
        ],
    )
    def test_create_role_all_privs_positive(self, priv_name, privs):
        """
        Create role positive
        """

        role_name = "usr-sys-admin-test-" + priv_name

        try:
            self.client.admin_get_role(role_name)
            # role exists, clear it out.
            self.client.admin_drop_role(role_name)
            time.sleep(2)
        except e.InvalidRole:
            pass  # we are good, no such role exists

        self.client.admin_create_role(role_name, privs, {"timeout": 1000})
        time.sleep(1)
        roles = self.client.admin_get_role(role_name)
        assert roles == {"privileges": privs, "whitelist": [], "read_quota": 0, "write_quota": 0}

        try:
            status = self.client.admin_create_user("testcreaterole", "createrole", [role_name])
        except e.QuotasNotEnabled:
            pytest.mark.skip(reason="Got QuotasNotEnabled, skipping quota test.")
            pytest.skip()

        assert status == 0
        time.sleep(1)
        roles = self.client.admin_query_user("testcreaterole")

        assert roles == [role_name]

        self.client.admin_drop_user("testcreaterole")

        self.client.admin_drop_role(role_name)

    def test_create_role_positive_with_policy_write(self):
        """
        Create role with write privilege positive
        """
        try:
            self.client.admin_get_role("usr-sys-admin-test")
            # role exists, clear it out.
            self.client.admin_drop_role("usr-sys-admin-test")
            time.sleep(2)
        except e.InvalidRole:
            pass  # we are good, no such role exists

        self.client.admin_create_role(
            "usr-sys-admin-test", [{"code": aerospike.PRIV_WRITE, "ns": "test", "set": "demo"}], {"timeout": 1000}
        )
        time.sleep(1)
        roles = self.client.admin_get_role("usr-sys-admin-test")
        assert roles == {
            "privileges": [{"code": 13, "ns": "test", "set": "demo"}],
            "whitelist": [],
            "read_quota": 0,
            "write_quota": 0,
        }

        try:
            status = self.client.admin_create_user("testcreaterole", "createrole", ["usr-sys-admin-test"])
        except e.QuotasNotEnabled:
            pytest.mark.skip(reason="Got QuotasNotEnabled, skipping quota test.")
            pytest.skip()

        assert status == 0
        time.sleep(1)
        roles = self.client.admin_query_user("testcreaterole")

        assert roles == ["usr-sys-admin-test"]

        self.client.admin_drop_user("testcreaterole")

    def test_create_role_positive(self):
        """
        Create role positive
        """
        try:
            self.client.admin_get_role("usr-sys-admin-test")
            # role exists, clear it out.
            self.client.admin_drop_role("usr-sys-admin-test")
            # Give some time for the role removal to take place
            time.sleep(2)
        except e.InvalidRole:
            pass  # we are good, no such role exists

        self.client.admin_create_role(
            "usr-sys-admin-test", [{"code": aerospike.PRIV_USER_ADMIN}, {"code": aerospike.PRIV_SYS_ADMIN}]
        )
        time.sleep(1)
        roles = self.client.admin_get_role("usr-sys-admin-test")

        assert roles == {
            "privileges": [{"code": 0, "ns": "", "set": ""}, {"code": 1, "ns": "", "set": ""}],
            "whitelist": [],
            "read_quota": 0,
            "write_quota": 0,
        }

        self.client.admin_drop_role("usr-sys-admin-test")

    def test_create_role_whitelist_quota_positive(self):
        """
        Create role positive
        """
        try:
            self.client.admin_get_role("usr-sys-admin-test")
            # role exists, clear it out.
            self.client.admin_drop_role("usr-sys-admin-test")
            # Give some time for the role removal to take place
            time.sleep(2)
        except e.InvalidRole:
            pass  # we are good, no such role exists

        try:
            self.client.admin_create_role(
                "usr-sys-admin-test",
                [{"code": aerospike.PRIV_USER_ADMIN}, {"code": aerospike.PRIV_SYS_ADMIN}],
                whitelist=["127.0.0.1", "10.1.2.0/24"],
                read_quota=20,
                write_quota=30,
            )
        except e.QuotasNotEnabled:
            pytest.mark.skip(reason="Got QuotasNotEnabled, skipping quota test.")
            pytest.skip()

        time.sleep(1)
        roles = self.client.admin_get_role("usr-sys-admin-test")

        assert roles == {
            "privileges": [
                {"code": aerospike.PRIV_USER_ADMIN, "ns": "", "set": ""},
                {"code": aerospike.PRIV_SYS_ADMIN, "ns": "", "set": ""},
            ],
            "whitelist": ["127.0.0.1", "10.1.2.0/24"],
            "read_quota": 20,
            "write_quota": 30,
        }

        self.client.admin_drop_role("usr-sys-admin-test")

    def test_create_role_incorrect_role_type(self):
        """
        role name not string
        """
        try:
            self.client.admin_create_role(1, [{"code": aerospike.PRIV_USER_ADMIN}])
        except e.ParamError as exception:
            assert exception.code == -2
            assert "Role name should be a string" in exception.msg

    def test_create_role_unknown_privilege_type(self):
        """
        privilege type unknown
        """
        try:
            self.client.admin_get_role("usr-sys-admin-test")
            # role exists, clear it out.
            self.client.admin_drop_role("usr-sys-admin-test")
            time.sleep(2)
        except e.InvalidRole:
            pass  # we are good, no such role exists

        try:
            self.client.admin_create_role("usr-sys-admin-test", [{"code": 64}])
        except e.InvalidPrivilege as exception:
            assert exception.code == 72

    def test_create_role_incorrect_privilege_type(self):
        """
        privilege type incorrect
        """
        try:
            self.client.admin_create_role("usr-sys-admin-test", None)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Privileges should be a list"

    def test_create_role_existing_role(self):
        """
        create an already existing role
        """
        try:
            self.client.admin_get_role("usr-sys-admin-test")
            # role exists, clear it out.
            self.client.admin_drop_role("usr-sys-admin-test")
            time.sleep(2)
        except e.InvalidRole:
            pass  # we are good, no such role exists

        self.client.admin_create_role(
            "usr-sys-admin-test", [{"code": aerospike.PRIV_USER_ADMIN}, {"code": aerospike.PRIV_SYS_ADMIN}]
        )
        try:
            self.client.admin_create_role(
                "usr-sys-admin-test", [{"code": aerospike.PRIV_USER_ADMIN}, {"code": aerospike.PRIV_SYS_ADMIN}]
            )

        except e.RoleExistsError as exception:
            assert exception.code == 71
            assert exception.msg == "AEROSPIKE_ROLE_ALREADY_EXISTS"

        time.sleep(1)
        status = self.client.admin_drop_role("usr-sys-admin-test")

        assert status == 0

    def test_create_role_positive_with_special_characters(self):
        """
        Create role positive with special characters in role name
        """
        role_name = "!#Q#AEQ@#$%&^*((^&*~~~````"
        try:
            self.client.admin_drop_role(role_name)  # clear out if it exists
            time.sleep(2)
        except Exception:
            pass  # EAFP
        status = self.client.admin_create_role(
            role_name, [{"code": aerospike.PRIV_READ, "ns": "test", "set": "demo"}], {"timeout": 1000}
        )

        assert status == 0
        time.sleep(1)
        roles = self.client.admin_get_role(role_name)

        assert roles == {
            "privileges": [
                {"code": aerospike.PRIV_READ, "ns": "test", "set": "demo"},
            ],
            "whitelist": [],
            "read_quota": 0,
            "write_quota": 0,
        }

        status = self.client.admin_create_user("testcreaterole", "createrole", [role_name])

        assert status == 0
        time.sleep(1)
        roles = self.client.admin_query_user("testcreaterole")

        assert roles == [role_name]

        self.client.admin_drop_role(role_name)

        time.sleep(1)

        roles = self.client.admin_query_user("testcreaterole")

        assert roles == []

        self.client.admin_drop_user("testcreaterole")

    def test_create_role_positive_with_too_long_role_name(self):
        """
        Create role positive with too long role name
        """
        role_name = "role$" * 1000

        try:
            self.client.admin_create_role(
                role_name, [{"code": aerospike.PRIV_READ, "ns": "test", "set": "demo"}], {"timeout": 1000}
            )

        except e.InvalidRole as exception:
            assert exception.code == 70
            assert exception.msg == "AEROSPIKE_INVALID_ROLE"



# -*- coding: utf-8 -*-
##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################

from __future__ import print_function

import aerospike
import signal
import sys
import time

# from guppy import hpy
from optparse import OptionParser

##########################################################################
# Options Parsing
##########################################################################

usage = "usage: %prog [options]"

optparser = OptionParser(usage=usage, add_help_option=False)

optparser.add_option(
    "--help", dest="help", action="store_true",
    help="Displays this message.")

optparser.add_option(
    "-U", "--username", dest="username", type="string", metavar="<USERNAME>",
    help="Username to connect to database.")

optparser.add_option(
    "-P", "--password", dest="password", type="string", metavar="<PASSWORD>",
    help="Password to connect to database.")

optparser.add_option(
    "-h", "--host", dest="host", type="string", default="127.0.0.1", metavar="<ADDRESS>",
    help="Address of Aerospike server.")

optparser.add_option(
    "-p", "--port", dest="port", type="int", default=3000, metavar="<PORT>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "-n", "--namespace", dest="namespace", type="string", default="test", metavar="<NS>",
    help="Namespace that records will be stored and retrieved from.")

optparser.add_option(
    "-s", "--set", dest="set", type="string", default="demo", metavar="<SET>",
    help="Set that records will be stored and retrieved from.")

optparser.add_option(
    "-v", "--verbose", dest="verbose", action="store_true", metavar="<PORT>",
    help="Verbose output.")


(options, args) = optparser.parse_args()

if options.help:
    optparser.print_help()
    print()
    sys.exit(1)

##########################################################################
# Client Configuration
##########################################################################

config = {
    'hosts': [(options.host, options.port)]
}

##########################################################################
# Application
##########################################################################

exitCode = 0

count = 0
start = 0


def total_summary():

    # stop time
    stop = time.time()

    # elapse time
    elapse = (stop - start)

    print()
    print("Summary:")
    print("     {0} keys generated".format(count))
    print("     {0} seconds for {1} operations".format(elapse, count))
    print("     {0} operations per second".format(count / elapse))
    print()
    # print("Heap: ")
    # print(heapy.heap())

    sys.exit(0)

try:
    # ----------------------------------------------------------------------------
    # Connect to Cluster
    # ----------------------------------------------------------------------------

    client = aerospike.client(config).connect(
        options.username, options.password)

    # ----------------------------------------------------------------------------
    # Perform Operation
    # ----------------------------------------------------------------------------

    try:

        signal.signal(signal.SIGTERM, total_summary)

        print()
        print("Press CTRL+C to quit.")
        print()

        start = time.time()

        # run the operatons
        while True:
            count += 1
            keyt = (options.namespace, options.set, count)
            client.put(keyt, {'key': count})

    except KeyboardInterrupt:
        total_summary()
    except Exception as eargs:
        print("error: {0}".format(eargs), file=sys.stderr)
        sys.exit(2)

except Exception as eargs:
    print("error: {0}".format(eargs), file=sys.stderr)
    sys.exit(3)

##########################################################################
# Exit
##########################################################################

sys.exit(0)



# -*- coding: utf-8 -*-
import pytest
from .test_base_class import TestBaseClass

import aerospike
from aerospike import exception as e


class TestOperateOrdered(object):
    def setup_class(cls):
        """
        Setup class.
        """
        cls.client_no_typechecks = TestBaseClass.get_new_connection({"strict_types": False})

    def teardown_class(cls):
        TestOperateOrdered.client_no_typechecks.close()

    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        keys = []

        for i in range(5):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "age": i}
            as_connection.put(key, rec)

        key = ("test", "demo", 6)
        rec = {"age": 6.3}
        as_connection.put(key, rec)
        keys.append(key)

        key = ("test", "demo", "bytearray_key")
        rec = {"bytearray_bin": bytearray("asd;as[d'as;d", "utf-8")}
        as_connection.put(key, rec)

        keys.append(key)

        key = ("test", "demo", "list_key")
        rec = {"int_bin": [1, 2, 3, 4], "string_bin": ["a", "b", "c", "d"]}
        as_connection.put(key, rec)

        keys.append(key)

        key = ("test", "demo", "existing_key")
        rec = {"dict": {"a": 1}, "bytearray": bytearray("abc", "utf-8"), "float": 3.4, "list": ["a"]}
        as_connection.put(key, rec)

        keys.append(key)

        def teardown():
            """
            Teardown Method
            """
            for i in range(5):
                key = ("test", "demo", i)
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    pass
            for key in keys:
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    pass

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "key, llist, expected",
        [
            (
                ("test", "demo", 1),
                [
                    {"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"},
                    {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
                    {"op": aerospike.OPERATOR_READ, "bin": "name"},
                ],
                [("name", "ramname1")],
            ),
            (
                ("test", "demo", 1),  # with_write_float_value
                [
                    {"op": aerospike.OPERATOR_WRITE, "bin": "write_bin", "val": {"no": 89.8}},
                    {"op": aerospike.OPERATOR_READ, "bin": "write_bin"},
                ],
                [("write_bin", {"no": 89.8})],
            ),
            (
                ("test", "demo", 1),  # write positive
                [
                    {"op": aerospike.OPERATOR_WRITE, "bin": "write_bin", "val": {"no": 89}},
                    {"op": aerospike.OPERATOR_READ, "bin": "write_bin"},
                ],
                [("write_bin", {"no": 89})],
            ),
            (
                ("test", "demo", 1),  # write_tuple_positive
                [
                    {"op": aerospike.OPERATOR_WRITE, "bin": "write_bin", "val": tuple("abc")},
                    {"op": aerospike.OPERATOR_READ, "bin": "write_bin"},
                ],
                [("write_bin", ("a", "b", "c"))],
            ),
            (
                ("test", "demo", 1),  # with_bin_bytearray
                [
                    {"op": aerospike.OPERATOR_PREPEND, "bin": bytearray("asd[;asjk", "utf-8"), "val": "ram"},
                    {"op": aerospike.OPERATOR_READ, "bin": bytearray("asd[;asjk", "utf-8")},
                ],
                [("asd[;asjk", "ram")],
            ),
            (
                ("test", "demo", "bytearray_key"),  # append_val bytearray
                [
                    {"op": aerospike.OPERATOR_APPEND, "bin": "bytearray_bin", "val": bytearray("abc", "utf-8")},
                    {"op": aerospike.OPERATOR_READ, "bin": "bytearray_bin"},
                ],
                [("bytearray_bin", bytearray("asd;as[d'as;dabc", "utf-8"))],
            ),
            (
                ("test", "demo", "bytearray_new"),  # append bytearray_newrecord
                [
                    {
                        "op": aerospike.OPERATOR_APPEND,
                        "bin": "bytearray_bin",
                        "val": bytearray("asd;as[d'as;d", "utf-8"),
                    },
                    {"op": aerospike.OPERATOR_READ, "bin": "bytearray_bin"},
                ],
                [("bytearray_bin", bytearray("asd;as[d'as;d", "utf-8"))],
            ),
            (
                ("test", "demo", "bytearray_key"),  # prepend_valbytearray
                [
                    {"op": aerospike.OPERATOR_PREPEND, "bin": "bytearray_bin", "val": bytearray("abc", "utf-8")},
                    {"op": aerospike.OPERATOR_READ, "bin": "bytearray_bin"},
                ],
                [("bytearray_bin", bytearray("abcasd;as[d'as;d", "utf-8"))],
            ),
            (
                ("test", "demo", "bytearray_new"),  # prepend_valbytearray_newrecord
                [
                    {
                        "op": aerospike.OPERATOR_PREPEND,
                        "bin": "bytearray_bin",
                        "val": bytearray("asd;as[d'as;d", "utf-8"),
                    },
                    {"op": aerospike.OPERATOR_READ, "bin": "bytearray_bin"},
                ],
                [("bytearray_bin", bytearray("asd;as[d'as;d", "utf-8"))],
            ),
        ],
    )
    def test_pos_operate_ordered_correct_paramters(self, key, llist, expected):
        """
        Invoke operate_ordered() with correct parameters
        """

        _, _, bins = self.as_connection.operate_ordered(key, llist)
        self.as_connection.remove(key)
        assert bins == expected

    def test_pos_operate_ordered_with_correct_policy(self):
        """
        Invoke operate_ordered() with correct policy
        """
        key = ("test", "demo", 1)
        policy = {
            "timeout": 180000,
            "key": aerospike.POLICY_KEY_SEND,
            "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
        }

        llist = [
            {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": "aa"},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
        ]

        _, _, bins = self.as_connection.operate_ordered(key, llist, {}, policy)
        self.as_connection.remove(key)
        assert bins == [("name", "name1aa")]

    def test_pos_operate_ordered_with_policy_key_digest(self):
        """
        Invoke operate_ordered() with policy_key_digest
        """
        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"))
        rec = {"name": "name%s" % (str(1)), "age": 1}
        policy = {"key": aerospike.POLICY_KEY_DIGEST}
        self.as_connection.put(key, rec)

        llist = [
            {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": "aa"},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "age"},
        ]

        _, _, bins = self.as_connection.operate_ordered(key, llist, {}, policy)

        assert bins == [("name", "name1aa"), ("age", 4)]
        self.as_connection.remove(key)

    @pytest.mark.parametrize(
        "key, policy, meta, llist",
        [
            (
                ("test", "demo", 1),
                {
                    "key": aerospike.POLICY_KEY_SEND,
                    "gen": aerospike.POLICY_GEN_IGNORE,
                    "commit_level": aerospike.POLICY_COMMIT_LEVEL_ALL,
                },
                {"gen": 10, "ttl": 1200},
                [
                    {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": "aa"},
                    {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
                    {"op": aerospike.OPERATOR_READ, "bin": "name"},
                ],
            ),
        ],
    )
    def test_pos_operate_ordered_policy_gen_ignore(self, key, policy, meta, llist):
        """
        Invoke operate_ordered() with gen ignore.
        """
        key, meta, bins = self.as_connection.operate_ordered(key, llist, meta, policy)

        assert bins == [("name", "name1aa")]

    def test_pos_operate_ordered_with_policy_gen_GT(self):
        """
        Invoke operate_ordered() with gen GT positive.
        """
        key = ("test", "demo", 1)
        policy = {"key": aerospike.POLICY_KEY_SEND, "gen": aerospike.POLICY_GEN_GT}
        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"]
        meta = {"gen": gen + 5, "ttl": 1200}

        llist = [
            {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": "aa"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        (key, meta, bins) = self.as_connection.operate_ordered(key, llist, meta, policy)

        assert bins == [("name", "name1aa")]

    def test_pos_operate_ordered_with_nonexistent_key(self):
        """
        Invoke operate_ordered() with non-existent key
        """
        key1 = ("test", "demo", "key11")
        llist = [
            {"op": aerospike.OPERATOR_PREPEND, "bin": "loc", "val": "mumbai"},
            {"op": aerospike.OPERATOR_READ, "bin": "loc"},
        ]
        _, _, bins = self.as_connection.operate_ordered(key1, llist)
        self.as_connection.remove(key1)
        assert bins == [("loc", "mumbai")]

    def test_pos_operate_ordered_increment_nonexistent_key(self):
        """
        Invoke operate_ordered() with increment with nonexistent_key
        """
        key = ("test", "demo", "non_existentkey")
        llist = [{"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 5}]

        _, _, op_bins = self.as_connection.operate_ordered(key, llist)
        _, _, get_bins = self.as_connection.get(key)
        self.as_connection.remove(key)

        assert op_bins == []
        assert get_bins == {"age": 5}

    def test_pos_operate_ordered_increment_nonexistent_bin(self):
        """
        Invoke operate_ordered() with increment with nonexistent_bin
        """
        key = ("test", "demo", 1)
        llist = [{"op": aerospike.OPERATOR_INCR, "bin": "my_age", "val": 5}]

        _, _, op_bins = self.as_connection.operate_ordered(key, llist)
        _, _, get_bins = self.as_connection.get(key)
        assert op_bins == []
        assert get_bins == {"my_age": 5, "age": 1, "name": "name1"}

    def test_pos_operate_ordered_write_set_to_aerospike_null(self):
        """
        Invoke operate_ordered() with write command bin set to aerospike_null
        """
        key = ("test", "demo", "null_record")

        bins = {"name": "John", "no": 3}

        assert 0 == self.as_connection.put(key, bins)

        llist = [
            {"op": aerospike.OPERATOR_WRITE, "bin": "no", "val": aerospike.null()},
            {"op": aerospike.OPERATOR_READ, "bin": "no"},
        ]

        _, _, bins = self.as_connection.operate_ordered(key, llist)
        self.as_connection.remove(key)
        # we read a non existent bin, so nothing is returned
        assert [] == bins

    # List operation testcases
    @pytest.mark.parametrize(
        "list, expected",
        [
            (
                [
                    {"op": aerospike.OP_LIST_APPEND, "bin": "int_bin", "val": 7},
                    {"op": aerospike.OP_LIST_GET, "bin": "int_bin", "index": 4},
                ],
                [("int_bin", 5), ("int_bin", 7)],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_APPEND_ITEMS, "bin": "int_bin", "val": [7, 9]},
                    {"op": aerospike.OP_LIST_GET_RANGE, "bin": "int_bin", "index": 3, "val": 3},
                ],
                [("int_bin", 6), ("int_bin", [4, 7, 9])],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_INSERT, "bin": "int_bin", "val": 7, "index": 2},
                    {"op": aerospike.OP_LIST_POP, "bin": "int_bin", "index": 2},
                ],
                [("int_bin", 5), ("int_bin", 7)],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_INSERT_ITEMS, "bin": "int_bin", "val": [7, 9], "index": 2},
                    {"op": aerospike.OP_LIST_POP_RANGE, "bin": "int_bin", "index": 2, "val": 2},
                ],
                [("int_bin", 6), ("int_bin", [7, 9])],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_SET, "bin": "int_bin", "index": 2, "val": 18},
                    {"op": aerospike.OP_LIST_GET, "bin": "int_bin", "index": 2},
                ],
                [("int_bin", 18)],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_SET, "bin": "int_bin", "index": 6, "val": 10},
                    {"op": aerospike.OP_LIST_GET, "bin": "int_bin", "index": 6},
                ],
                [("int_bin", 10)],
            ),
        ],
    )
    def test_pos_operate_ordered_with_list_addition_operations(self, list, expected):
        """
        Invoke operate_ordered() with list addition operations
        """
        key = ("test", "demo", "list_key")

        key, _, bins = self.as_connection.operate_ordered(key, list)

        assert bins == expected

    @pytest.mark.parametrize(
        "list, expected",
        [
            (
                [
                    {"op": aerospike.OP_LIST_REMOVE, "bin": "int_bin", "index": 2},
                ],
                [("int_bin", 1)],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_REMOVE_RANGE, "bin": "int_bin", "index": 2, "val": 2},
                ],
                [("int_bin", 2)],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_TRIM, "bin": "int_bin", "index": 2, "val": 2},
                ],
                [("int_bin", 2)],
            ),
            ([{"op": aerospike.OP_LIST_CLEAR, "bin": "int_bin"}], []),
        ],
    )
    def test_pos_operate_ordered_with_list_remove_operations(self, list, expected):
        """
        Invoke operate_ordered() with list remove operations
        """
        key = ("test", "demo", "list_key")

        _, _, bins = self.as_connection.operate_ordered(key, list)

        assert bins == expected

    def test_pos_operate_ordered_with_list_size(self):
        """
        Invoke operate_ordered() with list_size operation
        """
        key = ("test", "demo", "list_key")
        list = [{"op": aerospike.OP_LIST_SIZE, "bin": "int_bin"}]

        key, _, bins = self.as_connection.operate_ordered(key, list)

        assert bins == [("int_bin", 4)]

    def test_pos_operate_ordered_with_list_get_range_val_out_of_bounds(self):
        """
        Invoke operate_ordered() list_get_range op and value out of bounds
        """
        key = ("test", "demo", "list_key")
        list = [{"op": aerospike.OP_LIST_GET_RANGE, "bin": "int_bin", "index": 2, "val": 9}]

        (key, meta, bins) = self.as_connection.operate_ordered(key, list)

        assert bins == [("int_bin", [3, 4])]

    def test_pos_operate_ordered_with_list_trim_val_with_negative_value(self):
        """
        Invoke operate_ordered() with list_trimoperation and value is negative
        """
        key = ("test", "demo", "list_key")
        list = [{"op": aerospike.OP_LIST_TRIM, "bin": "int_bin", "index": 1, "val": -9}]

        (key, meta, bins) = self.as_connection.operate_ordered(key, list)

        assert bins == [("int_bin", 1)]

        (key, meta, bins) = self.as_connection.get(key)

        assert bins["int_bin"] == [2, 3, 4]

    def test_pos_operate_ordered_with_list_insert_index_negative(self):
        """
        Invoke operate_ordered()-list_insert and item index is a negative value
        """
        key = ("test", "demo", "list_key")
        list = [{"op": aerospike.OP_LIST_INSERT, "bin": "int_bin", "index": -2, "val": 9}]

        (key, meta, bins) = self.as_connection.operate_ordered(key, list)

        assert bins == [("int_bin", 5)]

        (key, meta, bins) = self.as_connection.get(key)

        assert bins["int_bin"] == [1, 2, 9, 3, 4]

    @pytest.mark.parametrize(
        "list, expected",
        [
            (
                [
                    {"op": aerospike.OP_LIST_APPEND, "bin": "string_bin", "val": {"new_val": 1}},
                    {"op": aerospike.OP_LIST_GET, "bin": "string_bin", "index": 4},
                ],
                [("string_bin", 5), ("string_bin", {"new_val": 1})],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_APPEND_ITEMS, "bin": "string_bin", "val": [["z", "x"], ("y", "w")]},
                    {"op": aerospike.OP_LIST_GET_RANGE, "bin": "string_bin", "index": 3, "val": 3},
                ],
                [("string_bin", 6), ("string_bin", ["d", ["z", "x"], ("y", "w")])],
            ),
            (
                [
                    {"op": aerospike.OP_LIST_INSERT, "bin": "string_bin", "val": True, "index": 2},
                    {"op": aerospike.OP_LIST_POP, "bin": "string_bin", "index": 2},
                ],
                [("string_bin", 5), ("string_bin", True)],
            ),
            (
                [
                    {
                        "op": aerospike.OP_LIST_INSERT_ITEMS,
                        "bin": "string_bin",
                        "val": [bytearray("abc", "utf-8"), "xyz"],
                        "index": 2,
                    },
                    {"op": aerospike.OP_LIST_POP_RANGE, "bin": "string_bin", "index": 2, "val": 2},
                ],
                [("string_bin", 6), ("string_bin", [bytearray(b"abc"), "xyz"])],
            ),
        ],
    )
    def test_pos_operate_ordered_with_list_ops_different_datatypes(self, list, expected):
        """
        Invoke operate_ordered() with list operations using different datatypes
        """
        key = ("test", "demo", "list_key")

        key, _, bins = self.as_connection.operate_ordered(key, list)

        assert bins == expected

    # No typecheck test cases
    @pytest.mark.parametrize(
        "key, llist, expected",
        [
            (
                ("test", "demo", "prepend_int"),  # prepend_with_int
                [
                    {"op": aerospike.OPERATOR_PREPEND, "bin": "age", "val": 4},
                    {"op": aerospike.OPERATOR_READ, "bin": "age"},
                ],
                [("age", 4)],
            ),
            (
                ("test", "demo", "append_dict"),  # append_with_dict
                [
                    {"op": aerospike.OPERATOR_APPEND, "bin": "dict", "val": {"a": 1, "b": 2}},
                    {"op": aerospike.OPERATOR_READ, "bin": "dict"},
                ],
                [("dict", {"a": 1, "b": 2})],
            ),
            (
                ("test", "demo", "incr_string"),  # incr_with_string
                [
                    {"op": aerospike.OPERATOR_INCR, "bin": "name", "val": "aerospike"},
                    {"op": aerospike.OPERATOR_READ, "bin": "name"},
                ],
                [("name", "aerospike")],
            ),
        ],
    )
    def test_pos_operate_ordered_new_record(self, key, llist, expected):
        """
        Invoke operate_ordered() with prepend command on a new record
        """
        try:
            self.as_connection.remove(key)
        except Exception:
            pass
        _, _, bins = TestOperateOrdered.client_no_typechecks.operate_ordered(key, llist)
        assert bins == expected

        try:
            self.as_connection.remove(key)
        except Exception:
            pass

    def test_pos_operate_ordered_with_bin_length_extra_nostricttypes(self):
        """
        Invoke operate_ordered() with bin length extra. Strict types disabled
        """
        key = ("test", "demo", 1)

        max_length = "a" * 21

        llist = [
            {"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"},
            {"op": aerospike.OPERATOR_INCR, "bin": max_length, "val": 3},
        ]

        _, _, bins = TestOperateOrdered.client_no_typechecks.operate_ordered(key, llist)

        assert bins == []

        (key, _, bins) = TestOperateOrdered.client_no_typechecks.get(key)

        assert bins == {"name": "ramname1", "age": 1}

    def test_pos_operate_ordered_with_command_invalid_nostricttypes(self):
        """
        Invoke operate_ordered() with an invalid command. Strict types disabled
        """
        key = ("test", "demo", 1)

        llist = [
            {"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"},
            {"op": 3, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        _, _, bins = TestOperateOrdered.client_no_typechecks.operate_ordered(key, llist)

        assert bins == [("name", "ramname1")]

    # Negative tests
    def test_neg_operate_ordered_with_no_parameters(self):
        """
        Invoke opearte_ordered() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.operate_ordered()
        assert "argument 'key' (pos 1)" in str(typeError.value)

    def test_neg_operate_ordered_with_policy_gen_EQ_not_equal(self):
        """
        Invoke operate_ordered() with gen not equal.
        """
        key = ("test", "demo", 1)
        policy = {"key": aerospike.POLICY_KEY_SEND, "gen": aerospike.POLICY_GEN_EQ}

        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"]
        meta = {"gen": gen + 5, "ttl": 1200}
        llist = [
            {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": "aa"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]
        try:
            key, meta, _ = self.as_connection.operate_ordered(key, llist, meta, policy)

        except e.RecordGenerationError as exception:
            assert exception.code == 3

        (key, meta, bins) = self.as_connection.get(key)
        assert bins == {"age": 1, "name": "name1"}
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_neg_operate_ordered_with_policy_gen_GT_lesser(self):
        """
        Invoke operate_ordered() with gen GT lesser.
        """
        key = ("test", "demo", 1)
        policy = {"key": aerospike.POLICY_KEY_SEND, "gen": aerospike.POLICY_GEN_GT}
        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"]
        meta = {"gen": gen, "ttl": 1200}

        llist = [
            {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": "aa"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        try:
            (key, meta, _) = self.as_connection.operate_ordered(key, llist, meta, policy)

        except e.RecordGenerationError as exception:
            assert exception.code == 3

        (key, meta, bins) = self.as_connection.get(key)
        assert bins == {"age": 1, "name": "name1"}
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_neg_operate_ordered_without_connection(self):
        """
        Invoke operate_ordered() with correct parameters without connection
        """
        key = ("test", "demo", 1)
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()
        llist = [
            {"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        try:
            key, _, _ = client1.operate_ordered(key, llist)

        except e.ClusterError as exception:
            assert exception.code == 11

    def test_neg_operate_ordered_prepend_set_to_aerospike_null(self):
        """
        Invoke operate_ordered() with prepend command bin set to aerospike_null
        """
        key = ("test", "demo", "null_record")

        bins = {"name": "John", "no": 3}

        assert 0 == self.as_connection.put(key, bins)

        llist = [
            {"op": aerospike.OPERATOR_PREPEND, "bin": "no", "val": aerospike.null()},
            {"op": aerospike.OPERATOR_READ, "bin": "no"},
        ]

        try:
            (key, _, bins) = self.as_connection.operate_ordered(key, llist)

        except e.InvalidRequest as exception:
            assert exception.code == 4
        self.as_connection.remove(key)

    def test_neg_operate_ordered_with_command_invalid(self):
        """
        Invoke operate_ordered() with an invalid command. Strict types enabled
        """
        key = ("test", "demo", 1)

        llist = [{"op": 3, "bin": "age", "val": 3}, {"op": aerospike.OPERATOR_READ, "bin": "name"}]

        try:
            key, _, _ = self.as_connection.operate_ordered(key, llist)

        except e.ParamError as exception:
            assert exception.code == -2

    def test_neg_operate_ordered_with_bin_length_extra(self):
        """
        Invoke operate_ordered() with bin length extra. Strict types enabled
        """
        key = ("test", "demo", 1)

        max_length = "a"
        for _ in range(20):
            max_length = max_length + "a"

        llist = [
            {"op": aerospike.OPERATOR_INCR, "bin": max_length, "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        try:
            key, _, _ = self.as_connection.operate_ordered(key, llist)

        except e.BinNameError as exception:
            assert exception.code == 21

    def test_neg_operate_ordered_empty_string_key(self):
        """
        Invoke operate_ordered() with empty string key
        """
        llist = [{"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"}]
        try:
            self.as_connection.operate_ordered("", llist)

        except e.ParamError as exception:
            assert exception.code == -2

    def test_neg_operate_ordered_with_extra_parameter(self):
        """
        Invoke operate_ordered() with extra parameter.
        """
        key = ("test", "demo", 1)
        policy = {}
        llist = [{"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"}]
        with pytest.raises(TypeError) as typeError:
            self.as_connection.operate_ordered(key, llist, {}, policy, "")

        assert "operate_ordered() takes at most 4 arguments (5 given)" in str(typeError.value)

    def test_neg_operate_ordered_policy_is_string(self):
        """
        Invoke operate_ordered() with policy is string
        """
        key = ("test", "demo", 1)
        llist = [{"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"}]
        try:
            self.as_connection.operate_ordered(key, llist, {}, "")

        except e.ParamError as exception:
            assert exception.code == -2

    def test_neg_operate_ordered_key_is_none(self):
        """
        Invoke operate_ordered() with key is none
        """
        llist = [{"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"}]
        try:
            self.as_connection.operate_ordered(None, llist)

        except e.ParamError as exception:
            assert exception.code == -2

    @pytest.mark.parametrize(
        "key, policy, list, ex_code",
        [
            (
                ("test", "demo", 1),
                {},
                [
                    {"op": aerospike.OPERATOR_APPEND, "bin": "name"},
                    {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
                ],
                -2,
            ),
            (
                ("test", "demo", 1),
                {},
                [
                    {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": 3, "aa": 89},
                ],
                -2,
            ),
            (
                ("test", "demo", 1),  # with_incr_value_string
                {
                    "key": aerospike.POLICY_KEY_SEND,
                    "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
                },
                [
                    {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": "3"},
                    {"op": aerospike.OPERATOR_READ, "bin": "age"},
                ],
                -2,
            ),
        ],
    )
    def test_neg_operate_ordered_append_without_value_parameter(self, key, policy, list, ex_code):
        """
        Invoke operate_ordered() with append op and append val is not given
        """

        try:
            self.as_connection.operate_ordered(key, list, {}, policy)

        except e.ParamError as exception:
            assert exception.code == ex_code

    def test_neg_operate_ordered_append_value_integer(self):
        """
        Invoke operate_ordered() with append value is of type integer
        """
        key = ("test", "demo", 1)
        llist = [
            {"op": aerospike.OPERATOR_APPEND, "bin": "name", "val": 12},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        try:
            self.as_connection.operate_ordered(key, llist)
        except e.ParamError as exc:
            assert exc.code == -2

    def test_neg_operate_ordered_with_incorrect_policy(self):
        """
        Invoke operate_ordered() with incorrect policy
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 0.5}
        llist = [
            {"op": aerospike.OPERATOR_PREPEND, "bin": "name", "val": "ram"},
            {"op": aerospike.OPERATOR_INCR, "bin": "age", "val": 3},
            {"op": aerospike.OPERATOR_READ, "bin": "name"},
        ]

        try:
            self.as_connection.operate_ordered(key, llist, {}, policy)

        except e.ParamError as exception:
            assert exception.code == -2

    def test_neg_operate_ordered_list_operation_bin_notlist(self):
        """
        Invoke operate_ordered() with a list op and bin does not contain list
        """
        key = ("test", "demo", 1)
        list = [{"op": aerospike.OP_LIST_INSERT, "bin": "age", "index": 2, "val": 9}]

        try:
            (key, _, _) = self.as_connection.operate_ordered(key, list)

        except e.BinIncompatibleType as exception:
            assert exception.code == 12

    def test_neg_operate_ordered_append_items_not_a_list(self):
        """
        Invoke operate_ordered() with list addition operations negative
        """
        key = ("test", "demo", "list_key")

        list = [
            {"op": aerospike.OP_LIST_APPEND_ITEMS, "bin": "int_bin", "val": 7},
        ]

        try:
            key, _, bins = self.as_connection.operate_ordered(key, list)

        except e.ParamError as exception:
            assert exception.code == -2

    @pytest.mark.parametrize(
        "key, llist",
        [
            (
                ("test", "demo", 1),
                [
                    {"op": aerospike.OPERATOR_PREPEND, "bin": "age", "val": 4},  # int
                    {"op": aerospike.OPERATOR_READ, "bin": "age"},
                ],
            ),
            (
                ("test", "demo", "existing_key"),  # Existing list
                [
                    {"op": aerospike.OPERATOR_PREPEND, "bin": "list", "val": ["c"]},
                    {"op": aerospike.OPERATOR_READ, "bin": "list"},
                ],
            ),
            (
                ("test", "demo", "existing_key"),  # Existing dict
                [
                    {"op": aerospike.OPERATOR_APPEND, "bin": "dict", "val": {"c": 2}},
                    {"op": aerospike.OPERATOR_READ, "bin": "dict"},
                ],
            ),
            (
                ("test", "demo", "existing_key"),  # Exiting float
                [
                    {"op": aerospike.OPERATOR_APPEND, "bin": "float", "val": 3.4},
                    {"op": aerospike.OPERATOR_READ, "bin": "float"},
                ],
            ),
            (
                ("test", "demo", 1),  # Existing string
                [
                    {"op": aerospike.OPERATOR_INCR, "bin": "name", "val": "aerospike"},
                    {"op": aerospike.OPERATOR_READ, "bin": "name"},
                ],
            ),
            (
                ("test", "demo", "existing_key"),  # Existing Bytearray
                [
                    {"op": aerospike.OPERATOR_INCR, "bin": "bytearray", "val": bytearray("abc", "utf-8")},
                    {"op": aerospike.OPERATOR_READ, "bin": "bytearray"},
                ],
            ),
        ],
    )
    def test_neg_operate_ordered_notypecheck_existing_record(self, key, llist):
        """
        Invoke operate() with no typecheck on existing record
        """
        try:
            (key, _, _) = TestOperateOrdered.client_no_typechecks.operate_ordered(key, llist)

        except e.BinIncompatibleType as exception:
            assert exception.code == 12

        TestOperateOrdered.client_no_typechecks.remove(key)



# -*- coding: utf-8 -*-

import pytest
import time
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


class TestGetRoles(TestBaseClass):

    pytestmark = pytest.mark.skipif(
        not TestBaseClass.auth_in_use(), reason="No user specified, may not be secured cluster."
    )

    def setup_method(self, method):
        """
        Setup method
        """
        config = TestBaseClass.get_connection_config()
        self.client = aerospike.client(config).connect(config["user"], config["password"])
        try:
            self.client.admin_drop_role("usr-sys-admin")
        except Exception:
            pass
        time.sleep(2)
        usr_sys_admin_privs = [{"code": aerospike.PRIV_USER_ADMIN}, {"code": aerospike.PRIV_SYS_ADMIN}]
        try:
            self.client.admin_drop_role("usr-sys-admin-test")
        except Exception:
            pass
        time.sleep(2)
        self.client.admin_create_role("usr-sys-admin-test", usr_sys_admin_privs)
        self.delete_users = []
        time.sleep(2)

    def teardown_method(self, method):
        """
        Teardown method
        """
        try:
            self.client.admin_drop_role("usr-sys-admin-test")
        except Exception:
            pass
        self.client.close()

    def test_admin_get_roles_positive(self):
        """
        Get roles positive
        """
        roles = self.client.admin_get_roles()

        assert roles["usr-sys-admin-test"] == {
            "privileges": [{"ns": "", "set": "", "code": 0}, {"ns": "", "set": "", "code": 1}],
            "whitelist": [],
            "read_quota": 0,
            "write_quota": 0,
        }

    def test_admin_get_roles_positive_with_policy(self):
        """
        Get roles positive policy
        """
        roles = self.client.admin_get_roles({"timeout": 1000})

        assert roles["usr-sys-admin-test"] == {
            "privileges": [{"ns": "", "set": "", "code": 0}, {"ns": "", "set": "", "code": 1}],
            "whitelist": [],
            "read_quota": 0,
            "write_quota": 0,
        }

    def test_admin_get_roles_incorrect_policy(self):
        """
        Get roles incorrect policy
        """
        try:
            self.client.admin_get_roles({"timeout": 0.2})

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "timeout is invalid"
